//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_cli.cpp
//	project: delib-cli
//
//
//	(c) DEDITEC GmbH, 2014
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// V1.00	DT	12.11.2013	Start DEDITEC-History
// V2.00	CC	14.01.2014	Added DX commands for TTL-I/O
//							Supports now FTDI linux driver (-> see "delib-cli-usb-ftdi")
// V2.01	CC	22.05.2014	Added DC/DCR command for digital counter
// V2.02	CC	25.06.2014	Set ADDA-Mode +/-10V if Command = AI/AO + V
//							Set ADDA-Mode 0..24nA if Command = AI/AO + mA
// V2.10	CC	07.09.2014	Added multiple interfaces (RO-ETH/LC, ETH-OPTOIN-8, ...)
// V2.20	CC	15.04.2016	Added x64 linux support
// V2.21	CC	26.07.2016	Added DC8/DCR8 commands to read multiple counter
//
#define verNr	0x0221					//Versionsnummer in Hex
//


// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// IsError
// check_delib_version_groesser_gleich
// print_usage
// chk_ch_count
// chk_argc_size
// read_config_file_and_get_params
// delib_cli_open_module

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdio.h>
#include <string.h>

#ifndef __unix__
	// ------ WINDOWS-----------
	// ------ WINDOWS-----------
	// ------ WINDOWS-----------
	#include <windows.h>
	#include "conio.h"

	#include "library/vc/deditec_string_convert_to_number.h"
	#include "library/vc/deditec_string_utils.h"
	#include "library/vc/delib_utilities1.h"
	#include "library/delib/delib.h"
	#include "library/delib/delib_extended_functions.h"
	#include "library/embedded/dt_var_defines.h"

	#include "delib_cli_defines.h"

	char global_ip_addr[20];

	// ------------------------
	// ------------------------
	// ------------------------
#else
	// ------ LINUX -----------
	// ------ LINUX -----------
	// ------ LINUX -----------

	#include "library/vc/deditec_string_convert_to_number.h"
	#include "library/vc/deditec_string_utils.h"
	#include "library/delib/delib.h"
	#include "library/embedded/dt_var_defines.h"

	#include "delib_cli_defines.h"

	extern ULONG read_config_file_and_get_params(unsigned long *buffer_module_id, unsigned long *buffer_module_nr);
	extern ULONG delib_cli_open_module(unsigned long moduleID, unsigned long moduleNR);
	
	// ------------------------
	// ------------------------
	// ------------------------
#endif

// ------------------------------------
// ------------------------------------
// Defines
// ------------------------------------
// ------------------------------------

//#define DEVELOPER_MODE 		1

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

#ifdef __unix__
	// ------ LINUX -----------
	// ------ LINUX -----------
	// ------ LINUX -----------
	bool IsError()
	{
		unsigned char msg[500];

		if (DapiGetLastError () != DAPI_ERR_NONE)
		{

			DapiGetLastErrorText((unsigned char*) msg, sizeof(msg));
			printf ("Error Code = %x * Message = %s\n", 0, msg);

			DapiClearLastError();

			return TRUE;
		}

		return FALSE;
	}

	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------

	ULONG check_delib_version_groesser_gleich(ULONG version)
	{
		unsigned int v0,v1,v2,v3;
		ULONG delib_version;
		
		delib_version = DapiGetDELIBVersion(0,0);

		if(delib_version < version)
		{
			v0=(version>>12) & 0xf;
			v1=(version>>8 ) & 0xf;
			v2=(version>>4 ) & 0xf;
			v3=(version>>0 ) & 0xf;

			if(v0==0)
			{
				printf("DELIB Version %x.%x%x or above is needed\n", v1, v2, v3);
			}
			else
			{
				printf("DELIB Version %x%x.%x%x or above is needed\n", v0, v1, v2, v3);
			}

			printf("please download at http://www.deditec.de/delib\n");
			printf("and install a newer driver\n");
			printf("\n");
			
			return RETURN_ERROR;
		}
		
		return RETURN_OK;
	}
	// ------------------------
	// ------------------------
	// ------------------------
#endif

// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------
// ---------------------------------------------------------------------------

void print_usage()
{
	printf("------------------------------------\n\r");
	printf("DELIB-CLI Version %d%d.%d%d for ", (verNr>>12&0xf), (verNr>>8&0xf), (verNr>>4&0xf), (verNr>>0&0xf));
	#ifdef  _WIN64
		printf("%d bit systems\n\r", 64);
	#else
		printf("%d bit systems\n\r", 32);
	#endif
	printf("------------------------------------\n\r");
	printf("\r\n");
	
	printf("------------------------------------\n\r");
	printf("------------------------------------\n\r");
	printf("USAGE\r\n");
	printf("------------------------------------\n\r");
	printf("DELIB-CLI is a utility to control several\r\n");
	printf("modules based on the DEDITEC Software Library (DELIB)\r\n");
	printf("\r\n");
	printf("delib_cli command channel [ value | unit [\"nounit\"] ]\r\n");
	printf("\r\n");

	printf("------------------------------------\n\r");
	printf("------------------------------------\n\r");
	printf("COMMANDS\r\n");
	printf("------------------------------------\n\r");
	printf("Digital Inputs:\t\tcommand = DI1, DI8, DI16, DI32, DI64\r\n");
	printf("Digital FlipFlops\tcommand = DF32\r\n");
	printf("Digital Counter:\tcommand = DC and DCR\r\n");
	printf("Digital Outputs:\tcommand = DO1, DO8, DO16, DO32, DO64\r\n");
	printf("TTL-I/O\t\t\tcommand = DX8\r\n");
	printf("Analog Inputs:\t\tcommand = AI\r\n");
	printf("Analog Outputs:\t\tcommand = AO\r\n");
	printf("------------------------------------\n\r");
	printf("channel:\t\ttrue numerical value\r\n");
	printf("value:\t\t\tyou can append following types to the value:\r\n");
	printf("\t\t\t0x(value) represents a hexadecimal value\r\n");
	printf("\t\t\t(value)mA represents a milli Ampere value\r\n");
	printf("\t\t\t(value)V represents a Volt value\r\n");
	printf("\t\t\tvalues without type will be parsed as decimal:\r\n");
	printf("\t\t\tif you add a floating number as value:\r\n");
	printf("\t\t\tyou are able to use \",\" or \".\" as comma\r\n");
	printf("unit:\t\t\tonly usable in case the command is DI or AI\r\n");
	printf("\t\t\twill define wich unit the printed value will have\r\n");
	printf("nounit:\t\t\tonly usable in case the command is DI or AI\r\n");
	printf("\t\t\targument is not case sensitive\r\n");
	printf("\r\n");
	
	printf("------------------------------------\n\r");
	printf("------------------------------------\n\r");
	printf("EXAMPLES\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli DO 17 1\r\n");
	printf("\tset digital output 17 to 1\r\n");
	printf("delib_cli DO1 3 0\r\n");
	printf("\tset digital output 3 to 0\r\n");
	printf("delib_cli DO8 0 5\r\n");
	printf("\tset digital output (0-7) to 5\r\n");
	printf("delib_cli DO16 0 0\r\n");
	printf("\tset digital output (0-15) to 0\r\n");
	printf("delib_cli DO16 16 2\r\n");
	printf("\tset digital output (16-31) to 2\r\n");
	printf("delib_cli DO32 0 7\r\n");
	printf("\tset digital output (0-31) to 7\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli DI 3\r\n");
	printf("\tread digital input 3\r\n");
	printf("delib_cli DI8 0\r\n");
	printf("\tread digital input (0-7)\r\n");
	printf("delib_cli DI16 16\r\n");
	printf("\tread digital input (16-31)\r\n");
	printf("delib_cli DI 3 hex\r\n");
	printf("\tread digital input 3 and return as hex\r\n");
	printf("delib_cli DI8 0 hex\r\n");
	printf("\tread digital input (0-7) and return as hex\r\n");
	printf("delib_cli DI16 16 hex\r\n");
	printf("\tread digital input (16-31) and return as hex\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli DF32 0\r\n");
	printf("\tread flipflops of digital input (0-31)\r\n");
	printf("delib_cli DF32 32\r\n");
	printf("\tread flipflops of digital input (32-63)\r\n");
	printf("delib_cli DF32 0 hex\r\n");
	printf("\tread flipflops of digital input (0-31) and return as hex\r\n");
	printf("delib_cli DF32 32 hex\r\n");
	printf("\tread flipflops of digital input (32-63) and return as hex\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli DC 4\r\n");
	printf("\tread digital counter 4\r\n");
	printf("delib_cli DCR 2\r\n");
	printf("\tread + reset digital counter 2\r\n");
	printf("delib_cli DC 3 hex\r\n");
	printf("\tread digital counter 3 and return as hex\r\n");
	printf("delib_cli DCR 21 hex\r\n");
	printf("\tread + reset digital counter 21 and return as hex\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli DX8 0 255\r\n");
	printf("\tset ttl-direction of ttl-io (0-63) to output\r\n");
	printf("delib_cli DX8 0 0\r\n");
	printf("\tset ttl-direction of ttl-io (0-63) to input\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli AO 7 4711\r\n");
	printf("\tset analog output no 7 to a decimal value\r\n");
	printf("delib_cli AO 7 0x4711\r\n");
	printf("\tset analog output no 7 to a hexadecimal value\r\n");
	printf("delib_cli AO 7 3.7V\r\n");
	printf("\tset analog output no 7 to a Volt value\r\n");
	printf("delib_cli AO 7 13.3mA\r\n");
	printf("\tset analog output no 7 to a milli Ampere value\r\n");
	printf("------------------------------------\n\r");
	printf("delib_cli AI 2\r\n");
	printf("\tread analog input no 2 as decimal value like: \"1234\"\r\n");
	printf("delib_cli AI 2 hex\r\n");
	printf("\tread analog input no 2 as hexadecimal value like: \"0x1FA\"\r\n");
	printf("delib_cli AI 2 V\r\n");
	printf("\tread analog input no 2 as Volt value like: \"3.7V\"\r\n");
	printf("delib_cli AI 2 mA\r\n");
	printf("\tread analog input no 2 as milli Ampere value like: \"13.3mA\"\r\n");
	printf("------------------------------------\n\r");
	printf("alternative \"nounit\" can be added to return typed values without type\r\n");
	printf("delib_cli AI 3 hex nounit\r\n");
	printf("\tread analog in no 3 as hexadecimal value like: \"1FA\"\r\n");
	printf("delib_cli AI 3 V nounit\r\n");
	printf("\tread analog in no 3 as Volt value like: \"3.7\"\r\n");
	printf("delib_cli AI 3 mA nounit\r\n");
	printf("\tread analog in no 3 as milli Ampere value like: \"13.3\"\r\n");
	printf("------------------------------------\n\r");
	printf("\n\r");
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long chk_ch_count (long ch, ULONG max, char cmd[])
{
	if (max == 0)
	{	
		if (strncmp(cmd, "DI", 2) == RETURN_OK)
		{
			printf("Error: Module has no digital inputs\r\n\r\n");
			return RETURN_ERROR;
		}
		else if (strncmp(cmd, "DO", 2) == RETURN_OK)
		{
			printf("Error: Module has no digital outputs\r\n\r\n");
			return RETURN_ERROR;
		}
		else if (strncmp(cmd, "DX", 2) == RETURN_OK)
		{
			printf("Error: Module has no TTL-IOs\r\n\r\n");
			return RETURN_ERROR;
		}
		else if (strncmp(cmd, "AI", 2) == RETURN_OK)
		{
			printf("Error: Module has no analog inputs\r\n\r\n");
			return RETURN_ERROR;
		}
		else if (strncmp(cmd, "AO", 2) == RETURN_OK)
		{
			printf("Error: Module has no analog outputs\r\n\r\n");
			return RETURN_ERROR;
		}
	}
	
	if ((ch < 0) || ((ULONG) ch >= max))
	{
		printf("Input Error: Channel shoud be in the range from 0 to %d\r\n\r\n", (unsigned int) max-1);
		print_usage();
		return RETURN_ERROR;
	}
	
	return RETURN_OK;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long chk_argc_size (ULONG typeofcmd, ULONG argc)
{
	switch (typeofcmd)
	{	
		// ------------------------------------
		case TYPE_MAIN:
			if ( (argc < 3) || (argc > 5) )
			{
				printf("input error: The counts of arguments should be between 2 and 4\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_DI:
			if ( (argc < 3) || (argc > 5) )
			{
				printf("input error: command=DI -> The counts of arguments should be between 1 and 3\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_DF:
			if ( (argc < 3) || (argc > 5) )
			{
				printf("input error: command=DF -> The counts of arguments should be between 1 and 3\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_DC:
			if ( (argc < 3) || (argc > 5) )
			{
				printf("input error: command=DC -> The counts of arguments should be between 1 and 3\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_DO:
			if (argc != 4)
			{
				printf("input error: command=DO -> The counts of arguments should be 2\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_DX:
			if (argc != 4)
			{
				printf("input error: command=DX -> The counts of arguments should be 2\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_AI:
			if ( (argc < 3) || (argc > 5) )
			{
				printf("input error: command=AI -> The counts of arguments should be between 1 and 3\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		case TYPE_AO:
			if (argc != 4)
			{
				printf("input error: command=AO -> The counts of arguments should be 2\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
			break;
		// ------------------------------------
		default:
			print_usage();
			return RETURN_ERROR;
		// ------------------------------------
	}
	
	return RETURN_OK;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifndef __unix__
	// ------ WINDOWS-----------
	// ------ WINDOWS-----------
	// ------ WINDOWS-----------

	unsigned long read_config_file_and_get_params(unsigned long *buffer_module_id, unsigned long *buffer_module_nr)
	{
		FILE *cfg_file;
		char text_buffer[100];
		char *text_content = NULL;
		char *param = NULL;
		BOOL found_roEth = FALSE;
		BOOL found_moduleID = FALSE;
		BOOL found_moduleNR = FALSE;
		unsigned long ret;
		
		cfg_file = fopen(CLI_CFG_FILE, "r");		// open cfg-file
		
		ret = 0;
		
		if(cfg_file == NULL) 	// can not open cfg-file
		{
	        printf("ERROR: Can not open 'delib_cli.cfg'\n");
		}
		else
		{
			ret |= RET_CFG_FILE_CFG_FILE_EXIST;
			
			while(!feof(cfg_file)) 		// end of file
			{
				if(fscanf(cfg_file, "%s", text_buffer) == EOF) 	
				{
					break;
				} 
				else 
				{
					param = strtok( text_buffer, "=");				// a=b;	-> a
					text_content = strtok( NULL, "=");				// a=b; -> b;
					text_content = strtok( text_content, ";");		// a=b; -> b
					
					if (strcmp(param, MODULE_ID) == 0)
					{
						*buffer_module_id = (unsigned long) atoi(text_content);
						found_moduleID = TRUE;
					}
					else if (strcmp(param, MODULE_NR) == 0)
					{
						*buffer_module_nr = (unsigned long) atoi(text_content);
						found_moduleNR = TRUE;
					}
					else if (strcmp(param, RO_ETH_IP) == 0)
					{
						sprintf((char*)global_ip_addr, "%s", text_content);
						found_roEth = TRUE;
					}
				}
			}		
		}
		

		fclose(cfg_file);
		
		if (found_roEth == TRUE)
		{
			ret |= RET_CFG_FILE_IP_FOUND;
		}
		
		if (found_moduleNR == TRUE)
		{
			ret |= RET_CFG_FILE_MODULE_NR_FOUND;
		}
		
		if (found_moduleID == TRUE)
		{
			ret |= RET_CFG_FILE_MODULE_ID_FOUND;
		}
		
		if ((found_moduleNR == FALSE) || (found_moduleID == FALSE))
		{
			printf("ERROR: No entry for moduleID and/or moduleNR found in 'delib_cli.cfg'\n");
		}
		else
		{
			switch(*buffer_module_id)
			{
				// USB
				case RO_USB1:
				case USB_MINI_STICK:
				case USB_OPTOIN_8:
				case USB_OPTOIN_8_RELAIS_8:
				case USB_OPTOIN_16_RELAIS_16:
				case USB_OPTOIN_32_RELAIS_32:
				case USB_TTL_32:
				case RO_USB2:

				// ETH
				case RO_ETH:
				case BS_ETH:
				case RO_ETH_LC:
				case ETH_RELAIS_8:
				case ETH_TTL_64:

					ret |= RET_CFG_FILE_MODULE_ID_IS_VALID;
					break;
					
				default:
					printf("ERROR: Value of moduleID in 'delib_cli.cfg' is not valid\n");
					printf("List of supported moduleIDs\n");
					printf("*** USB ***\n");
					printf("RO-USB=7\n");
					printf("USB-MINI-Stick=9\n");
					printf("USB-OPTOIN-8=14\n");
					printf("USB-RELAIS-8=14\n");
					printf("BS-USB-OPTOIN-8_RELAIS_8=15\n");
					printf("BS-USB-OPTOIN-16_RELAIS_16=16\n");
					printf("BS-USB-OPTOIN-32=16\n");
					printf("BS-USB-RELAIS-32=16\n");
					printf("BS-USB-OPTOIN-32_RELAIS_32=17\n");
					printf("BS-USB-OPTOIN-64=17\n");
					printf("BS-USB-RELAIS-64=17\n");
					printf("USB-TTL-32=18\n");
					printf("USB-TTL-64=18\n");
					printf("RO-USB2=25\n");
					printf("*** ETH ***\n");
					printf("RO-ETH=8\n");
					printf("BS-ETH=22\n");
					printf("RO-ETH/LC=26\n");
					printf("ETH-OPTOIN-8=27\n");
					printf("ETH-RELAIS-8=27\n");
					printf("ETH-TTL-32=30\n");
					printf("ETH-TTL-64=30\n");
					break;
			}
		}
		
		return ret;
	}

	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------

	ULONG delib_cli_open_module(unsigned long moduleID, unsigned long moduleNR, unsigned long ret_of_read_config_file_function)
	{
		ULONG handle = 0;
		ULONG interface_type = 0;
		DAPI_OPENMODULEEX_STRUCT open_buffer;
		ULONG mask = 0;

		
		if ((interface_type = DapiInternCommand(DAPI_INTERN_GET_MODULE_INTERFACE_TYPE, "", "", "", 0, &moduleID)) == DAPI_INTERFACE_UNDEFINED)
		{
			// error
			return 0;
		}
			
		if (interface_type != DAPI_INTERFACE_ETH)
		{
			// No ETH
			mask = (RET_CFG_FILE_MODULE_NR_FOUND | RET_CFG_FILE_MODULE_ID_FOUND | RET_CFG_FILE_MODULE_ID_IS_VALID);	// valid id + nr found
			if ((ret_of_read_config_file_function & mask) == mask)
			{
				handle = DapiOpenModule(moduleID, moduleNR);
			}
		}
		else
		{
			// ETH
			if ((ret_of_read_config_file_function & RET_CFG_FILE_IP_FOUND) == RET_CFG_FILE_IP_FOUND)	// ip found -> DapiOpenModuleEx
			{
				strcpy_n((char*) open_buffer.address, sizeof(open_buffer.address), global_ip_addr, strlen_n(global_ip_addr, sizeof(global_ip_addr)));
				open_buffer.timeout = 1000;			// 1000 msec
				open_buffer.portno = 0;				// using default port
				open_buffer.encryption_type = DAPI_OPEN_MODULE_ENCRYPTION_TYPE_NONE;

				handle = DapiOpenModuleEx(ETHERNET_MODULE, 0, (unsigned char*) &open_buffer, DAPI_OPEN_MODULE_OPTION_USE_EXBUFFER);
			}
			else	// open Eth via DapiOpenModule with windows registry values
			{
				handle = DapiOpenModule(moduleID, moduleNR);
			}
		}

		return handle;
	}

	// ------------------------
	// ------------------------
	// ------------------------
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
	ULONG par1, par2;
	ULONG handle;
	unsigned char par2_nounit;
	long num_conv_ret_pos;
	float par2_float;
	ULONG cmd_length;
	char cmd[256];
	ULONG par2_length;
	char par2_char[256];
	par2_nounit = 0;
	
	ULONG moduleID;
	ULONG moduleNR;
	ULONG ret;
	
	ULONG no_of_di;
	ULONG no_of_do;
	ULONG no_of_dx;
	ULONG no_of_ai;
	ULONG no_of_ao;
	
	unsigned char buffer[256];
	ULONG i;
	
	#ifndef DEVELOPER_MODE
		if(check_delib_version_groesser_gleich(0x135) != RETURN_OK) 
		{
			return RETURN_ERROR;
		}
	#endif

	if (chk_argc_size(TYPE_MAIN, argc) != RETURN_OK) 
	{
		return RETURN_ERROR;
	}

	#ifdef DEVELOPER_MODE
		int i = 1;
		printf("\r\ncmd:");
		while (i < argc)
		{
			printf(" %s", argv[i]);
			i++;
		}
		printf("\r\n");
		handle = 1;
	#else
		ret = read_config_file_and_get_params(&moduleID, &moduleNR);
	
		#ifdef __unix__
			handle = delib_cli_open_module(moduleID, moduleNR);
		#else
			handle = delib_cli_open_module(moduleID, moduleNR, ret);
		#endif
		
		if (handle == 0)
		{
			printf("No module found!\n");
			return RETURN_ERROR;
		}
	#endif
	
	if (handle != 0)
	{
		#ifdef DEVELOPER_MODE
			no_of_dx = 4,
			no_of_di = 32,
			no_of_do = 32,
			no_of_ai = 8,
			no_of_ao = 8;
		#else
			no_of_dx = DapiSpecialCommand(handle, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG, DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DX, 0, 0);
			IsError();
			no_of_di = DapiSpecialCommand(handle, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG, DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DI, 0, 0);
			IsError();
			no_of_do = DapiSpecialCommand(handle, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG, DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DO, 0, 0);
			IsError();
			no_of_ai = DapiSpecialCommand(handle, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG, DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_AD, 0, 0);
			IsError();
			no_of_ao = DapiSpecialCommand(handle, DAPI_SPECIAL_CMD_GET_MODULE_CONFIG, DAPI_SPECIAL_GET_MODULE_CONFIG_PAR_DA, 0, 0);
			IsError();
		#endif

		cmd_length = strlen_n(argv[1], -1);
		strcpy_n (cmd, 256, argv[1], cmd_length);
		toupperstr_n(cmd, strlen_n(cmd, sizeof(cmd)));

		par1 = dt_stringtoulong(argv[2], strlen_n(argv[2], -1), &num_conv_ret_pos);
		if ( (num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != strlen_n(argv[2], -1)) )
		{
			printf("input error: argument 2 is invalid\r\n\r\n");
			print_usage();
			return RETURN_ERROR;
		}

		if (argc >= 4)
		{
			par2_length = strlen_n(argv[3], -1);
			strcpy_n (par2_char, 256, argv[3], par2_length);
			toupperstr_n(par2_char, strlen_n(par2_char, sizeof(par2_char)));
		}

		if (argc == 5)
		{
			if (strlen_n(argv[4], -1) == 6)
			{
				char nounit[7];
				strcpy_n(nounit, 7, argv[4], 6);
				toupperstr_n(nounit, strlen_n(nounit, sizeof(nounit)));
				if(strcmp(nounit, "NOUNIT") == RETURN_OK)
				{
					par2_nounit = 1;
				}
				else
				{
					printf("input error: argument 4 is invalid\r\n\r\n");
					print_usage();
					return RETURN_ERROR;
				}
			}
			else
			{
				printf("input error: argument 4 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
		}

		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		if(strncmp(cmd, "DI", 2) == RETURN_OK)
		{
			// Command = DI
			if((chk_argc_size (TYPE_DI, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_di, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}
			
			if (argc >= 4)
			{
				// ------------------------------------
				if (strcmp(par2_char, "HEX") == RETURN_OK)
				{
					// ------------------------------------
					if ((strcmp(cmd, "DI") == RETURN_OK) || (strcmp(cmd, "DI1") == RETURN_OK))
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGet1 (ULONG %lu, ULONG %lu); returns: %s%X\r\n", handle, par1, (par2_nounit ? "" : "0x"), 123);
						#else
							printf("%s%X\n", (par2_nounit ? "" : "0x"), (unsigned int)DapiDIGet1(handle, par1));
						#endif
					}
					// ------------------------------------
					else if (strcmp(cmd, "DI8") == RETURN_OK)
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGet8 (ULONG %lu, ULONG %lu); returns: %s%X\r\n", handle, par1, (par2_nounit ? "" : "0x"), 123);
						#else
							printf("%s%X\n", (par2_nounit ? "" : "0x"), (unsigned int)DapiDIGet8(handle, par1));
						#endif
					}
					// ------------------------------------
					else if (strcmp(cmd, "DI16") == RETURN_OK)
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGet16 (ULONG %lu, ULONG %lu); returns: %s%X\r\n", handle, par1, (par2_nounit ? "" : "0x"), 123);
						#else
							printf("%s%X\n", (par2_nounit ? "" : "0x"), (unsigned int)DapiDIGet16(handle, par1));
						#endif
					}
					// ------------------------------------
					else if (strcmp(cmd, "DI32") == RETURN_OK)
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGet32 (ULONG %lu, ULONG %lu); returns: %s%X\r\n", handle, par1, (par2_nounit ? "" : "0x"), 123);
						#else
							printf("%s%X\n", (par2_nounit ? "" : "0x"), (unsigned int)DapiDIGet32(handle, par1));
						#endif
					}
					// ------------------------------------
					else
					{
						printf("input error: argument 1 is invalid\r\n\r\n");
						print_usage();
						return RETURN_ERROR;
					}
				}
				// ------------------------------------
				else
				{
					printf("input error: argument 3 is invalid\r\n\r\n");
					print_usage();
					return RETURN_ERROR;
				}
			}
			// ------------------------------------
			else if ((strcmp(cmd, "DI") == RETURN_OK) || (strcmp(cmd, "DI1") == RETURN_OK))
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGet1 (ULONG %lu, ULONG %lu); returns: %d\r\n", handle, par1, 123);
				#else
					unsigned long ret = DapiDIGet1(handle, par1);
					printf("%lu\n", ret);
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DI8") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGet8 (ULONG %lu, ULONG %lu); returns: %d\r\n", handle, par1, 123);
				#else
					printf("%lu\n", DapiDIGet8(handle, par1));
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DI16") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGet16 (ULONG %lu, ULONG %lu); returns: %d\r\n", handle, par1, 123);
				#else
					printf("%lu\n", DapiDIGet16(handle, par1));
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DI32") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGet32 (ULONG %lu, ULONG %lu); returns: %d\r\n", handle, par1, 123);
				#else
					printf("%lu\n", DapiDIGet32(handle, par1));
				#endif
			}
			else
			{
				printf("input error: argument 1 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else if(strncmp(cmd, "DF", 2) == RETURN_OK)
		{
			// Command = DF32
			
			if ((chk_argc_size(TYPE_DF, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_di, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}	
			
			if (argc >= 4)
			{
				if (strcmp(par2_char, "HEX") == RETURN_OK)
				{
					if (strcmp(cmd, "DF32") == RETURN_OK)
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGetFF32 (ULONG %lu, ULONG %lu); returns: %s%X\r\n", handle, par1, (par2_nounit ? "" : "0x"), 123);
						#else
							printf("%s%X\n", (par2_nounit ? "" : "0x"), (unsigned int)DapiDIGetFF32(handle, par1));
						#endif
					}
					else
					{
						printf("input error: argument 1 is invalid\r\n\r\n");
						print_usage();
						return RETURN_ERROR;
					}
				}	
				else
				{
					printf("input error: argument 3 is invalid\r\n\r\n");
					print_usage();
					return RETURN_ERROR;
				}
			}
			// ------------------------------------
			else if (strcmp(cmd, "DF32") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGetFF32 (ULONG %lu, ULONG %lu); returns: %d\r\n", handle, par1, 123);
				#else
					unsigned long ret = DapiDIGetFF32(handle, par1);
					IsError();
					printf("%lu\n", ret);
				#endif
			}
			else
			{
				printf("input error: argument 1 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else if(strncmp(cmd, "DC", 2) == RETURN_OK)
		{
			// Command = DC
			
			if ((chk_argc_size(TYPE_DF, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_di, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}
			
			if (argc >= 4)
			{
				if (strcmp(par2_char, "HEX") == RETURN_OK)
				{

					if (strcmp(cmd, "DC") == RETURN_OK)
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGetCounter (ULONG %lu, ULONG %lu, ULONG 0); returns: 123\r\n", handle, par1);
						#else
							ret = DapiDIGetCounter(handle, par1, 0);
							IsError();
							printf("%s%lX\n", (par2_nounit ? "" : "0x"), ret);
						#endif
					}
					// ------------------------------------
					else if (strcmp(cmd, "DCR") == RETURN_OK)
					{
						#ifdef DEVELOPER_MODE
							printf("DapiDIGetCounter (ULONG %lu, ULONG %lu, ULONG 1); returns: 123\r\n", handle, par1);
						#else
							ret = DapiDIGetCounter(handle, par1, 1);
							IsError();
							printf("%s%lX\n", (par2_nounit ? "" : "0x"), ret);
						#endif
					}
					else if (strcmp(cmd, "DC8") == RETURN_OK)
					{
						ret = DapiReadMultipleBytes(handle, 0x100 + par1*2, 16, 1, buffer, 16);
						IsError();
						
						printf("[");
						for (i=0; i!=16; i+=2)
						{
							ret  = (buffer[i+0] << 0);
							ret |= (buffer[i+1] << 8);
							printf("0x%lx", ret);
							if (i!=14)
							{
								printf(", ");
							}
						}
						printf("]\n");
					}
					// ------------------------------------
					else if (strcmp(cmd, "DCR8") == RETURN_OK)
					{
						ret = DapiReadMultipleBytes(handle, 0x200 + par1*2, 16, 1, buffer, 16);
						IsError();
						
						printf("[");
						for (i=0; i!=16; i+=2)
						{
							ret  = (buffer[i+0] << 0);
							ret |= (buffer[i+1] << 8);
							printf("0x%lx", ret);
							if (i!=14)
							{
								printf(", ");
							}
						}
						printf("]\n");
					}
					else
					{
						printf("input error: argument 1 is invalid\r\n\r\n");
						print_usage();
						return RETURN_ERROR;
					}
				}	
				else
				{
					printf("input error: argument 3 is invalid\r\n\r\n");
					print_usage();
					return RETURN_ERROR;
				}
			}
			// ------------------------------------
			else if (strcmp(cmd, "DC") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGetCounter (ULONG %lu, ULONG %lu, ULONG 0); returns: 123\r\n", handle, par1);
				#else
					unsigned long ret = DapiDIGetCounter(handle, par1, 0);
					IsError();
					printf("%lu\n", ret);
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DCR") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDIGetCounter (ULONG %lu, ULONG %lu, ULONG 1); returns: 123\r\n", handle, par1);
				#else
					unsigned long ret = DapiDIGetCounter(handle, par1, 1);
					IsError();
					printf("%lu\n", ret);
				#endif
			}
			else if (strcmp(cmd, "DC8") == RETURN_OK)
			{
				ret = DapiReadMultipleBytes(handle, 0x100 + par1*2, 16, 1, buffer, 16);
				IsError();
				
				printf("[");
				for (i=0; i!=16; i+=2)
				{
					ret  = (buffer[i+0] << 0);
					ret |= (buffer[i+1] << 8);
					printf("%lu", ret);
					if (i!=14)
					{
						printf(", ");
					}
				}
				printf("]\n");
			}
			// ------------------------------------
			else if (strcmp(cmd, "DCR8") == RETURN_OK)
			{
				ret = DapiReadMultipleBytes(handle, 0x200 + par1*2, 16, 1, buffer, 16);
				IsError();
				
				printf("[");
				for (i=0; i!=16; i+=2)
				{
					ret  = (buffer[i+0] << 0);
					ret |= (buffer[i+1] << 8);
					printf("%lu", ret);
					if (i!=14)
					{
						printf(", ");
					}
				}
				printf("]\n");
			}
			
			else
			{
				printf("input error: argument 1 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else if (strncmp(cmd, "DO", 2) == RETURN_OK)
		{
			if ((chk_argc_size(TYPE_DO, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_do, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}
			
			if (par2_length > 2)
			{
				if (strncmp(par2_char, "0X", 2) == RETURN_OK)
				{
					par2 = dt_hexstringtoulong((unsigned char*)(par2_char + 2), par2_length - 2, (unsigned long*) &num_conv_ret_pos);
				}
				else
				{
					par2 = dt_stringtoulong(par2_char, par2_length, &num_conv_ret_pos);
				}
			}
			else
			{
				par2 = dt_stringtoulong(par2_char, par2_length, &num_conv_ret_pos);
			}

			if ((num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != par2_length))
			{
				printf("input error: argument 2 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}

			// ------------------------------------
			if ((strcmp(cmd, "DO") == RETURN_OK) || (strcmp(cmd, "DO1") == RETURN_OK))
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDOSet1 (ULONG %lu, ULONG %lu, ULONG %lu);\n", handle, par1, par2);
				#else
					DapiDOSet1(handle, par1, par2);
					IsError();
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DO8") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDOSet8 (ULONG %lu, ULONG %lu, ULONG %lu);\n", handle, par1, par2);
				#else
					DapiDOSet8(handle, par1, par2);
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DO16") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDOSet16 (ULONG %lu, ULONG %lu, ULONG %lu);\n", handle, par1, par2);
				#else
					DapiDOSet16(handle, par1, par2);
				#endif
			}
			// ------------------------------------
			else if (strcmp(cmd, "DO32") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiDOSet32 (ULONG %lu, ULONG %lu, ULONG %lu);\n", handle, par1, par2);
				#else
					DapiDOSet32(handle, par1, par2);
				#endif
			}
			// ------------------------------------
			else
			{
				printf("input error: argument 1 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else if (strncmp(cmd, "DX", 2)  == RETURN_OK)
		{
			if ((chk_argc_size(TYPE_DX, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_dx, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}

			if (par2_length > 2)
			{
				if (strncmp(par2_char, "0X", 2) == RETURN_OK)
				{
					par2 = dt_hexstringtoulong((unsigned char*) (par2_char + 2), par2_length - 2, (unsigned long*) &num_conv_ret_pos);
				}
				else
				{
					par2 = dt_stringtoulong(par2_char, par2_length, &num_conv_ret_pos);
				}
			}
			else
			{
				par2 = dt_stringtoulong(par2_char, par2_length, &num_conv_ret_pos);
			}

			if ((num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != par2_length))
			{
				printf("input error: argument 2 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}

			if (strcmp(cmd, "DX8") == RETURN_OK)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiSpecialCommand (ULONG %lu, DAPI_SPECIAL_CMD_SET_DIR_DX_8, ULONG %lu, ULONG %lu, 0);\n", handle, par1, par2);
				#else
					DapiSpecialCommand(handle, DAPI_SPECIAL_CMD_SET_DIR_DX_8, par1, par2, 0);
					IsError();
				#endif
			}
			else
			{
				printf("input error: argument 1 is invalid\r\n\r\n");
				print_usage();
				return RETURN_ERROR;
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else if (strcmp(cmd, "AO") == RETURN_OK)
		{
			// Commmand = AO
			if ((chk_argc_size (TYPE_AO, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_ao, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}

			while(1)
			{
				// ------------------------------------
				// VOLT
				if (par2_length > 1)
				{
					if (((char*) par2_char)[par2_length -1] == 'V')
					{
						par2_float = dt_stringtofloat(par2_char, par2_length - 1, &num_conv_ret_pos);

						if ((num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != par2_length - 1))
						{
							printf("input error: argument 2 is invalid\r\n\r\n");
							print_usage();
							return RETURN_ERROR;
						}

						#ifdef DEVELOPER_MODE
							printf("DapiDaSetVolt (ULONG %lu, ULONG %lu, float %f);\r\n", handle, par1, par2_float);
						#else
							DapiDASetMode(handle, par1, DAPI_ADDA_MODE_BIPOL_10V);
							DapiDASetVolt(handle, par1, par2_float);
						#endif
						break;
					}
				}

				if (par2_length > 2)
				{
					// ------------------------------------
					//MA 
					if (strncmp((((char*) par2_char) +par2_length -2), "MA", 2) == RETURN_OK)
					{
						par2_float = dt_stringtofloat(par2_char, par2_length - 2, &num_conv_ret_pos);

						if ((num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != par2_length - 2))
						{
							printf("input error: argument 2 is invalid\r\n\r\n");
							print_usage();
							return RETURN_ERROR;
						}

						#ifdef DEVELOPER_MODE
							printf("DapiDaSetmA (ULONG %lu, ULONG %lu, float %f);\r\n", handle, par1, par2_float);
						#else
							DapiDASetMode(handle, par1, DAPI_ADDA_MODE_0_24mA);
							DapiDASetmA(handle, par1, par2_float);
						#endif
						break;
					}
					// ------------------------------------
					// HEX
					if (strncmp(par2_char, "0X", 2) == RETURN_OK)
					{
						par2 = dt_hexstringtoulong((unsigned char*) (par2_char + 2), par2_length - 2, (unsigned long*) &num_conv_ret_pos);
						if ((num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != (par2_length - 2)))
						{
							printf("input error: argument 2 is invalid\r\n\r\n");
							print_usage();
							return RETURN_ERROR;
						}
					}
					// ------------------------------------
					else
					{
						//decimal
						par2 = dt_stringtoulong(par2_char, par2_length, &num_conv_ret_pos);
						if ((num_conv_ret_pos == -1) || ((ULONG) num_conv_ret_pos != par2_length))
						{
							printf("input error: argument 2 is invalid\r\n\r\n");
							print_usage();
							return RETURN_ERROR;
						}
					}
				// ------------------------------------

					#ifdef DEVELOPER_MODE
						printf("DapiDaSet (ULONG %lu, ULONG %lu, ULONG %lu);\r\n", handle, par1, par2);
					#else
						DapiDASet(handle, par1, par2);
					#endif
				}
				break;
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else if (strcmp(cmd, "AI") == RETURN_OK)
		{
			// Commmand = AI

			if ((chk_argc_size (TYPE_AI, argc) != RETURN_OK) || (chk_ch_count(par1, no_of_ai, cmd) != RETURN_OK))
			{
				return RETURN_ERROR;
			}

			if (argc >= 4)
			{
				// ------------------------------------
				if (strcmp(par2_char, "HEX") == RETURN_OK)
				{
					#ifdef DEVELOPER_MODE
						printf("DapiADGet (ULONG %lu, ULONG %lu); returns: %s%X\r\n", handle, par1, (par2_nounit ? "" : "0x"), 123);
					#else
						printf("%s%X\n", (par2_nounit ? "" : "0x"), (unsigned int)DapiADGet(handle, par1));
					#endif
				}
				// ------------------------------------
				else if (strcmp(par2_char, "MA") == RETURN_OK)
				{
					#ifdef DEVELOPER_MODE
						printf("DapiADGetmA (ULONG %lu, ULONG %lu); returns: %s\r\n", handle, par1, (par2_nounit ? "" : "mA"));
					#else
						DapiADSetMode(handle, par1, DAPI_ADDA_MODE_0_24mA);
						printf("%f%s\n", DapiADGetmA(handle, par1), (par2_nounit ? "" : "mA"));
					#endif
				}
				// ------------------------------------
				else if (strcmp(par2_char, "V") == RETURN_OK)
				{
					#ifdef DEVELOPER_MODE
						printf("DapiADGetVolt (ULONG %lu, ULONG %lu); returns: %s\r\n", handle, par1, (par2_nounit ? "" : "V"));
					#else
						DapiADSetMode(handle, par1, DAPI_ADDA_MODE_BIPOL_10V);
						printf("%f%s\n", DapiADGetVolt(handle, par1), (par2_nounit ? "" : "V"));
					#endif
				}
				// ------------------------------------
				else
				{
					printf("input error: command=AI -> argument: \"%s\" appears to be invalid\r\n\r\n", argv[3]);
					print_usage();
					return RETURN_ERROR;
				}
			}
				// ------------------------------------
			else if (argc == 3)
			{
				#ifdef DEVELOPER_MODE
					printf("DapiADGet (ULONG %lu, ULONG %lu); returns: %d\r\n", handle, par1, 123);
				#else
					printf("%lu\n", DapiADGet(handle, par1));
				#endif
			}
		}
		
		// ------------------------------------
		// ------------------------------------
		// ------------------------------------
		
		else
		{
			print_usage();
			return RETURN_ERROR;
		}
		
		#ifndef DEVELOPER_MODE
			//DapiCloseModule(handle);
		#endif
	}
	else
	{
		printf("ERROR: Module could not be opened\r\n");
	}
	
	return RETURN_OK;
}
