//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_io_ro_eth_intern.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiOpenModuleIntern_ETH_INTERN
// DapiCloseModuleIntern_ETH_INTERN
// ------------------------------------
// DapiROETHInternWriteByte
// DapiROETHInternWriteWord
// DapiROETHInternWriteLong
// DapiROETHInternWriteLongLong
// ------------------------------------
// DapiROETHInternReadByte
// DapiROETHInternReadWord
// DapiROETHInternReadLong
// DapiROETHInternReadLongLong
// ------------------------------------
// DapiROETHInternWriteMultipleBytes
// DapiROETHInternReadMultipleBytes

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdio.h>

#include "proj_config.h"
#include "delib_debug.h"	

#include "delib_linux_project_include-eth_intern.h"

#include "library/embedded/dt_embedded_parameter_linux.h"
#include "library/delib/delib.h"
#include "library/delib/delib_intern.h"

#include "library/linux/deditec_msgqueue.h"
#include "library/linux/deditec_msgqueue_struct.h"
#include "library/linux/deditec_msgq_srv_thread.h"
#include "library/linux/deditec_msgq_send_message_to_queue.h"

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiOpenModuleIntern_ETH_INTERN(ULONG handle, ULONG moduleID)
{
	char debug_msg[200];
	ULONG i;
	DAPI_HANDLE *DapiHandle;
	
   	DapiHandle = (DAPI_HANDLE*) handle;

	// DEBUG
	debug_set_function_name("DapiOpenModuleIntern_ETH_INTERN");
	sprintf(debug_msg,"DapiOpenModuleIntern_ETH_INTERN(handle=0x%lx, moduleID=0x%lx)", handle, moduleID);
	debug_info(debug_msg);
	// -----	
	
	if (DapiHandle->id != delib_id)
	{
		debug_set_function_name("DapiOpenModuleIntern_ETH_INTERN");
		sprintf(debug_msg,"Invalid Handle-ID");
		debug_error(debug_msg);
		return FALSE;
	}

	DapiHandle->moduleID = moduleID;
	DapiHandle->SubDevice = 2;
	
	return 0;	// -> ok
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiCloseModuleIntern_ETH_INTERN(ULONG handle)
{
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	
   	DapiHandle = (DAPI_HANDLE*) handle;

	if (handle == 0)
	{
		// DEBUG
		debug_set_function_name("DapiCloseModuleIntern_ETH_INTERN");
		sprintf(debug_msg, "Invalid Handle (handle=0)");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiCloseModuleIntern_ETH_INTERN");
		sprintf(debug_msg, "Invalid Handle-ID");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if(DapiCheckHandle(handle) != 0) 
	{
		return -1;
	}

	tcp_tx_close(DapiHandle->TCPHandle);
	
	DapiHandle->moduleID = -1;

	return TRUE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiROETHInternWriteByte(unsigned long handle, unsigned long adress, unsigned long value)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternWriteByte");
		sprintf(debug_msg,"DapiROETHInternWriteByte(handle=0x%lx, address=0x%lx, value=0x%lx", handle, adress, value);
		debug_print(debug_msg);
	#endif
	// -----	
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'W';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'W';								// write
	putdata.content.key[tx_cnt++] = 'B';								// byte
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	tx_cnt = 0;
	putdata.content.value[tx_cnt++] = (UCHAR) value;
	
	putdata.content.vallen = (UCHAR) tx_cnt;


	// ------------------------------------------------------------------------
	// an queue schicken

	if(msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_HW_IO, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiROETHInternWriteWord (unsigned long handle,unsigned long adress,unsigned long value)
{	
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternWriteWord");
		sprintf(debug_msg,"DapiROETHInternWriteWord(handle=0x%lx, address=0x%lx, value=0x%lx)", handle, adress, value);
		debug_print(debug_msg);
	#endif
	// -----	
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'W';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'W';								// write
	putdata.content.key[tx_cnt++] = 'W';								// word
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	tx_cnt = 0;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 8) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 0) & 0xff;
	
	putdata.content.vallen = (UCHAR) tx_cnt;


	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiROETHInternWriteLong (unsigned long handle,unsigned long adress,unsigned long value)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternWriteLong");
		sprintf(debug_msg,"DapiROETHInternWriteLong(handle=0x%lx, address=0x%lx, value=0x%lx)", handle, adress, value);
		debug_print(debug_msg);
	#endif
	// -----	
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'W';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'W';								// write
	putdata.content.key[tx_cnt++] = 'L';								// long
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	tx_cnt = 0;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 24) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 16) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 8) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 0) & 0xff;
	
	putdata.content.vallen = (UCHAR) tx_cnt;


	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


void DapiROETHInternWriteLongLong (unsigned long handle, unsigned long adress, ULONGLONG value)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternWriteLongLong");
		sprintf(debug_msg,"DapiROETHInternWriteLongLong(handle=0x%lx, address=0x%lx, value=0x%llx)", handle, adress, value);
		debug_print(debug_msg);
	#endif
	// -----
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'W';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'W';								// write
	putdata.content.key[tx_cnt++] = 'X';								// longlong
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	tx_cnt = 0;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 56) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 48) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 40) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 32) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 24) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 16) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 8) & 0xff;
	putdata.content.value[tx_cnt++] = (UCHAR) (value >> 0) & 0xff;
	
	putdata.content.vallen = (UCHAR) tx_cnt;

	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


unsigned long DapiROETHInternReadByte (unsigned long handle,unsigned long adress)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	ULONG data;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadByte");
		sprintf(debug_msg,"DapiROETHInternReadByte(handle=0x%lx, address=0x%lx)", handle, adress);
		debug_print(debug_msg);
	#endif
	// -----
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'R';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'R';								// read
	putdata.content.key[tx_cnt++] = 'B';								// byte
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	putdata.content.vallen = 0;


	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);	
		return(0L);		
	}
	else
	{
		data = (unsigned long) getdata.content.value[0];
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadByte");
		sprintf(debug_msg,"return = 0x%lx", data);
		debug_print(debug_msg);
	#endif
	// -----
	
	return data;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiROETHInternReadWord (unsigned long handle,unsigned long adress)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	ULONG data;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadWord");
		sprintf(debug_msg,"DapiROETHInternReadWord(handle=0x%lx, address=0x%lx)", handle, adress);
		debug_print(debug_msg);
	#endif
	// -----
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'R';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'R';								// read
	putdata.content.key[tx_cnt++] = 'W';								// word
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	putdata.content.vallen = 0;


	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);	
		return(0L);		
	}
	else
	{
		data =  ((unsigned long) (getdata.content.value[1]) << 8) & 0xff00;
		data |= ((unsigned long) (getdata.content.value[0]) << 0) & 0x00ff;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadWord");
		sprintf(debug_msg,"return = 0x%lx", data);
		debug_print(debug_msg);
	#endif
	// -----
	
	return data;
}
	
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


unsigned long DapiROETHInternReadLong(unsigned long handle, unsigned long adress)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	ULONG data;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadLong");
		sprintf(debug_msg,"DapiROETHInternReadLong(handle=0x%lx, address=0x%lx)", handle, adress);
		debug_print(debug_msg);
	#endif
	// -----
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'R';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'R';								// read
	putdata.content.key[tx_cnt++] = 'L';								// long
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	putdata.content.vallen = 0;


	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);	
		return(0L);		
	}
	else
	{
		data =  ((unsigned long) (getdata.content.value[3]) << 24) & 0xff000000;
		data |= ((unsigned long) (getdata.content.value[2]) << 16) & 0x00ff0000;
		data |= ((unsigned long) (getdata.content.value[1]) << 8 ) & 0x0000ff00;
		data |= ((unsigned long) (getdata.content.value[0]) << 0 ) & 0x000000ff;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadLong");
		sprintf(debug_msg,"return = 0x%lx", data);
		debug_print(debug_msg);
	#endif
	// -----
	
	return data;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


ULONGLONG DapiROETHInternReadLongLong(unsigned long handle, unsigned long adress)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	ULONGLONG data;
	ULONG data_L;
	ULONG data_H;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadLongLong");
		sprintf(debug_msg,"DapiROETHInternReadLongLong(handle=0x%lx, address=0x%lx)", handle, adress);
		debug_print(debug_msg);
	#endif
	// -----

	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'R';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';								// register mode
	putdata.content.key[tx_cnt++] = 'R';								// read
	putdata.content.key[tx_cnt++] = 'X';								// longlong
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 8) & 0xff);		// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((adress >> 0) & 0xff);		// addr_bit_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	putdata.content.vallen = 0;

	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);	
		return(0L);		
	}
	else
	{
		data_H =  ((unsigned long) (getdata.content.value[7]) << 24) & 0xff000000;
		data_H |= ((unsigned long) (getdata.content.value[6]) << 16) & 0x00ff0000;
		data_H |= ((unsigned long) (getdata.content.value[5]) << 8 ) & 0x0000ff00;
		data_H |= ((unsigned long) (getdata.content.value[4]) << 0 ) & 0x000000ff;
		
		data_L =  ((unsigned long) (getdata.content.value[3]) << 24) & 0xff000000;
		data_L |= ((unsigned long) (getdata.content.value[2]) << 16) & 0x00ff0000;
		data_L |= ((unsigned long) (getdata.content.value[1]) << 8 ) & 0x0000ff00;
		data_L |= ((unsigned long) (getdata.content.value[0]) << 0 ) & 0x000000ff;
		
		data = data_H;
		data = (data << 32) | data_L;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadLongLong");
		sprintf(debug_msg,"return = 0x%llx", data);
		debug_print(debug_msg);
	#endif
	// -----
	
	return data;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiROETHInternWriteMultipleBytes(ULONG handle, ULONG address, ULONG address_depth, ULONG increment, unsigned char *data, ULONG data_length)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;
	char debug_msg[200];

	ULONG tx_cnt;
	ULONG i;
	ULONG j;
	ULONG pos;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternWriteMultipleBytes");
		sprintf(debug_msg,"DapiROETHInternWriteMultipleBytes(handle=0x%lx, addresse=0x%lx, addr_depth=0x%lx, increment=0x%lx, data_length=0x%lx)", handle, address, address_depth, increment, data_length);
		debug_print(debug_msg);
	#endif
	// -----
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'W';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';										// register mode
	putdata.content.key[tx_cnt++] = 'W';										// write
	putdata.content.key[tx_cnt++] = 'M';										// multiple
	putdata.content.key[tx_cnt++] = (UCHAR) ((address >> 8) & 0xff);			// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((address >> 0) & 0xff);			// addr_bit_7_0
	putdata.content.key[tx_cnt++] = (UCHAR) ((address_depth >> 8) & 0xff);		// addr_depth_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((address_depth >> 0) & 0xff);		// addr_depth_7_0
	putdata.content.key[tx_cnt++] = (UCHAR) ((increment >> 8) & 0xff);			// increment_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((increment >> 0) & 0xff);			// increment_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;

	// value
	tx_cnt = 0;
	
	for (i=0; i<increment; i++)
	{
		// START_ID_MULTIPLE_BYTE (1byte)
		putdata.content.value[tx_cnt++]= DEDITEC_TCP_START_ID_FOR_MULTIPLE_BYTE_DATA;		// START_ID_MULTIPLE_BYTE
		
		// DATA_ID (2byte)
		putdata.content.value[tx_cnt++]= (unsigned char) ((i >> 8 ) & 0xff);			// DATA_ID_BIT_8_15
		putdata.content.value[tx_cnt++]= (unsigned char) ((i >> 0 ) & 0xff);			// DATA_ID_BIT_0_7
		
		// data
		for (j=0; j<address_depth; j++)
		{
			putdata.content.value[tx_cnt++]= (*(data + (pos++)));
		}
	}
	
	putdata.content.vallen = tx_cnt;

	// ------------------------------------------------------------------------
	// an queue schicken

	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);
		return 1;	// ERROR
	}
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiROETHInternReadMultipleBytes(ULONG handle, ULONG address, ULONG address_depth, ULONG increment, unsigned char *data, ULONG data_length)
{
	dt_msgq_c2s putdata;
	dt_msgq_s2c getdata;

	char debug_msg[200];
	ULONG tx_cnt;
	ULONG i;
	ULONG j;
	ULONG pos;
	
	ULONG m_data_pos;
	ULONG m_start_id;
	ULONG m_data_id;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiROETHInternReadMultipleBytes");
		sprintf(debug_msg,"DapiROETHInternReadMultipleBytes(handle=0x%lx, addresse=0x%lx, addr_depth=0x%lx, increment=0x%lx, data_length=0x%lx)", handle, address, address_depth, increment, data_length);
		debug_print(debug_msg);
	#endif
	// -----
	
	// ------------------------------------------------------------------------
	// check ob data_length gro genug ist
	
	if (data_length < (increment*address_depth))
	{
		//sprintf(debug_msg, "DapiTCPReadMultipleBytes / Buffer too small - address_depth=%ld * increment=%ld < data_length=%ld", address_depth, increment, data_length);
		//DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);
	}
	
	// ------------------------------------------------------------------------
	// queue nachricht zusammenbauen
	
	// cmd
	putdata.content.cmd = 'R';
	
	// key
	tx_cnt = 0;
	putdata.content.key[tx_cnt++] = 'R';										// register mode
	putdata.content.key[tx_cnt++] = 'R';										// read
	putdata.content.key[tx_cnt++] = 'M';										// multiple
	putdata.content.key[tx_cnt++] = (UCHAR) ((address >> 8) & 0xff);			// addr_bit_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((address >> 0) & 0xff);			// addr_bit_7_0
	putdata.content.key[tx_cnt++] = (UCHAR) ((address_depth >> 8) & 0xff);		// addr_depth_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((address_depth >> 0) & 0xff);		// addr_depth_7_0
	putdata.content.key[tx_cnt++] = (UCHAR) ((increment >> 8) & 0xff);			// increment_15_8
	putdata.content.key[tx_cnt++] = (UCHAR) ((increment >> 0) & 0xff);			// increment_7_0
	
	putdata.content.keylen = (UCHAR) tx_cnt;
	
	// value
	putdata.content.vallen = 0;
	
	// ------------------------------------------------------------------------
	// an queue schicken
	
	if(dt_msgq_send_message_to_queue_with_init_and_close(DT_MSGQ_RO_ETH_FW, &putdata, &getdata) != 0)	// error
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);	
		return(0L);		
	}
	else	// ok
	{
		// antwort berprfen
		for (i=0; i<increment; i++)
		{
			pos = 0;
			m_data_pos = i*(address_depth+3);
			
			// start_id + data_id aus dem buffer filtern
			m_start_id = getdata.content.value[m_data_pos + (pos++)];
			m_data_id = getdata.content.value[m_data_pos + (pos++)];
			m_data_id = (m_data_id << 8) | getdata.content.value[m_data_pos + (pos++)];
			
			// start_id checken
			if (m_start_id != DEDITEC_TCP_START_ID_FOR_MULTIPLE_BYTE_DATA)
			{
				//sprintf(debug_msg, "DapiTCPReadMultipleBytes / Wrong START_ID");
				//DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

				return 1;			// Error
			}
			
			// data_id checken
			if (m_data_id != i)
			{	
				//sprintf(debug_msg, "DapiTCPReadMultipleBytes / Wrong DATA_ID");
				//DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

				return 1;			// Error
			}
		}
		
		// daten sind ok - daten in "richtigen buffer" schreiben
		pos = 0;
		tx_cnt = 0;
		
		for (i=0; i<increment; i++)
		{
			for (j=0; j<address_depth; j++)
			{
				*(data + (tx_cnt++)) = getdata.content.value[pos + 3 + i];
			}
			
			pos += address_depth;
		}
	
	
		// geschriebene zeichen im "richtigen buffer" mit anzahl "soll-zeichen" vergleichen
		if (tx_cnt != (increment * address_depth))
		{
			//sprintf(debug_msg, "DapiTCPReadMultipleBytes / Buffer copy error");
			//DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

			return 1;			// Error
		}
	}
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
