//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_io_seriellFTDI.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiSeriellFTDISendCommand
// ------------------------------------
// DapiSeriellFTDIWriteByte
// DapiSeriellFTDIWriteWord
// DapiSeriellFTDIWriteLong
// DapiSeriellFTDIWriteLongLong
// DapiSeriellFTDIWrite64BytesMax
// ------------------------------------
// DapiSeriellFTDIReadByte
// DapiSeriellFTDIReadWord
// DapiSeriellFTDIReadLong
// DapiSeriellFTDIReadLongLong
// DapiSeriellFTDIRead64BytesMax

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdio.h>

#include "proj_config.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	
	#ifdef USE_USB_SIMPLE
		#include "../usb-simple/ftd2xx.h"
	#else
		#include "../usb-ftdi/ftd2xx.h"	
	#endif
	
	#include "library/delib/delib.h"
	#include "library/vc/deditec_tcp_io.h"	// muss vor die "delib_intern.h", weil hier TCP_IO drin ist
	#include "library/delib/delib_intern.h"
	#include "library/vc/deditec_convert_functions.h"
	#include "delib_packet_statistic.h"
	
	#include "delib_set_get_error.h"
	#include "delib_hdlmgr.h"
	#include "delib_io_ftdi.h"
	
	#include "library/embedded/deditec_debug_comfort.h"
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include <windows.h>
	#include "delib_io_ftdi.h"
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
#endif

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiSeriellFTDISendCommand(ULONG handle32, unsigned char tx[], unsigned long tx_length, unsigned char rx[], unsigned long rx_anzahl)
{
//	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	FT_STATUS ftStatus;
	unsigned long i;
	DWORD ret;
	ULONG anz_retry;
	unsigned char checksum;
	unsigned char rx_count;
	unsigned char rx_data[200];
	unsigned char rx_pos;
	unsigned long ok;
	unsigned char tx_buffer[200];
	unsigned char tx_buffer_cnt;

	unsigned char send_char;
	unsigned char modul_nr;
	ULONG rx_nr;
	
	char debug_msg[200];
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	// DEBUG
	debug_collect_init();

	debug_set_function_name("DapiSeriellFTDISendCommand");

	sprintf(debug_msg, "DapiSeriellFTDISendCommand START");
	debug_collect(debug_msg);

	sprintf(debug_msg, "DapiSeriellFTDISendCommand(tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, (unsigned int) tx[0],(unsigned int) tx[1], (unsigned int) rx_anzahl);
	debug_collect(debug_msg);
	// -----
	
	modul_nr = (unsigned char) DapiHandle->RS485_ModuleNr;
	++DapiHandle->jobID;
	
	// Immer eins mehr, da ja mindestens ein Mal bertragen werden muss
	anz_retry = DapiHandle->retry + 1;
		
	do 
	{
		if(anz_retry != (DapiHandle->retry+1))
		{
			// Empfangsbuffer flushen
			ret = FTread_n_bytes_with_timeout(handle32, 200, rx_data, 50*1000);	// Read x Byte (max 50ms->zum flushen)

			if(ret != 0)
			{
				sprintf(debug_msg, "FLUSH BUFFER - ret = %d bytes wurden geflusht", ret);
				debug_collect(debug_msg);
				if (ret > 0)
				{
					for (i=0; i<ret; i+=16)
					{
						debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), rx_data, sizeof(rx_data), i, ret);
						debug_collect(debug_msg);
					}
				}
			}
		}

		ok=0;
	
		// -----------------------------------------------------------
		// Ab hier wird die Checksumme berechnet
	
		checksum=0;

		tx_buffer_cnt=0;
		tx_buffer[tx_buffer_cnt++] = 0xfe;		// Dummy
		tx_buffer[tx_buffer_cnt++] = 0xfe;		// Dummy
		tx_buffer[tx_buffer_cnt++] = 0x01;		// SOH ASCII=StartOfHeader

		checksum+= 0x01;

		// Sende Modul-Nr - H
		send_char = (unsigned char) hex2asc((modul_nr >> 4) & 0x0f);
		tx_buffer[tx_buffer_cnt++] = send_char;
		checksum += send_char;

		// Sende Modul-Nr - L
		send_char = (unsigned char) hex2asc((modul_nr >> 0) & 0x0f);
		tx_buffer[tx_buffer_cnt++] = send_char;
		checksum += send_char;

		// Sende Job-ID - H
		send_char = (unsigned char) hex2asc((DapiHandle->jobID >> 4) & 0x0f);
		tx_buffer[tx_buffer_cnt++] = send_char;
		checksum += send_char;

		// Sende Job-ID - L
		send_char = (unsigned char) hex2asc((DapiHandle->jobID >> 0) & 0x0f);
		tx_buffer[tx_buffer_cnt++] = send_char;
		checksum += send_char;

		// Zeichen aus Sendestring dazu
		for(i=0;i<tx_length;++i)
		{
			checksum+=tx[i];
			tx_buffer[tx_buffer_cnt++] = tx[i];
		}

		// Checksumme dazu
		tx_buffer[tx_buffer_cnt++]=(unsigned char) hex2asc((checksum >> 4) & 0x0f);
		tx_buffer[tx_buffer_cnt++]=(unsigned char) hex2asc((checksum     ) & 0x0f);
	
		// Abschlusszeichen
		tx_buffer[tx_buffer_cnt++] = 13;		// Abschlusszeichen

		// DEBUG
		debug_set_function_name("DapiSeriellFTDISendCommand");
		sprintf(debug_msg, "jetzt senden");
		debug_collect(debug_msg);
		// -----

		tx_buffer[tx_buffer_cnt]=0;		// Terminate String

		// DEBUG
		debug_set_function_name("DapiSeriellFTDISendCommand");
		sprintf(debug_msg, "Sende string : (length=%d)= string=-->%s<--\r0", tx_buffer_cnt, tx_buffer);
		debug_collect(debug_msg);
		// -----
		
		// Alles an FTDI-Com senden
		ftStatus = FT_Write(DapiHandle->ftHandle, tx_buffer, tx_buffer_cnt, &ret);

		// DEBUG
		debug_set_function_name("DapiSeriellFTDISendCommand");
		sprintf(debug_msg, "send:ret=%d",(unsigned int) ret);
		debug_collect(debug_msg);
		sprintf(debug_msg, "Senden fertig (retry=%d)\r0",(unsigned int) anz_retry);
		debug_collect(debug_msg);
		// -----
		
		// -----------------------------------------------------------
		// Empfange jetzt einen Antwortstring (bis einschl. CR)

		rx_nr = rx_anzahl + 6; // 'D' + 2*Job_id + 2* Checksum + CR -> 6
		rx_count = (unsigned char) FTread_n_bytes_with_timeout(handle32, rx_nr, rx_data, 100*1000);	// Read 2 Byte (max 100ms)
		rx_data[rx_count]=0;		// Terminate String

		// DEBUG
		debug_set_function_name("DapiSeriellFTDISendCommand");
		sprintf(debug_msg, "Antwort (length=%d)= string=-->%s<--\r0", rx_count, rx_data);
		debug_collect(debug_msg);
		if (rx_count > 0)
		{
			for (i=0; i<rx_count; i+=16)
			{
				debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), rx_data, sizeof(rx_data), i, rx_count);
				debug_collect(debug_msg);
			}
		}
		// -----

		if(rx_count!=0)
		{
			if(rx_data[0]=='E')	// 'E' = Error
			{
				// Error
				if(rx_count>1)
				{
					// DEBUG
					debug_set_function_name("DapiSeriellFTDISendCommand");
					sprintf(debug_msg, "Received-Error Code = %c\r0", rx_data[1]);
					debug_collect(debug_msg);
					// -----

					switch(rx_data[1])
					{
						//case '1':	delib_error_code = DAPI_ERR_DEV_PKT_CMD_ILLEGAL;		break;
						//case '2':	delib_error_code = DAPI_ERR_DEV_PKT_DATA_LENGTH_ERROR;	break;
						//case '3':	delib_error_code = DAPI_ERR_DEV_PKT_DATA_CHECKSUM_ERROR;break;
						//case '4':	delib_error_code = DAPI_ERR_DEV_PKT_ADDR_NOT_FOR_ME;	break;
						case 'a':	return DAPI_ERR_DEV_IO_ADDR_ILLEGAL;		break;				// Nur bei diesen Error Codes soll abgebrochen werden !
						case 'b':	return DAPI_ERR_DEV_IO_RD_WR_ERROR;			break;				// Nur bei diesen Error Codes soll abgebrochen werden !
						//default:	delib_error_code = DAPI_ERR_DEV_PKT_OTHER_ERROR;		break;
					}

					//return (delib_error_code);	// DELIB ERROR
				}
			}
			if((rx_data[0]=='D') || (rx_data[0]=='O'))	// 'D' jetzt kommen Daten / 'O' Alles ok !
			{
				rx_pos=1;
				checksum=rx_data[0];

				// Job ID ok ??
				if(((unsigned char) hex2asc((DapiHandle->jobID >> 4) & 0x0f) == rx_data[rx_pos++]) && \
				   ((unsigned char) hex2asc((DapiHandle->jobID >> 0) & 0x0f) == rx_data[rx_pos++]))
				{
					// Job-ID ist ok !
					checksum+=rx_data[1];
					checksum+=rx_data[2];

					// Soll auch empfangen werden ???
					if(rx_anzahl!=0)	
					{
						for(i=0;i!=rx_anzahl;++i)
						{
							rx[i]=rx_data[rx_pos];	// Copy Data
							checksum+=rx_data[rx_pos];	// Add checksum

							// DEBUG
							debug_set_function_name("DapiSeriellFTDISendCommand");
							sprintf(debug_msg, "Daten Empfang: rx=0x%x  i=%d rx_pos=%d rx_anzahl=%d\r0", (unsigned char) rx_data[rx_pos], (unsigned int) i, (unsigned int) rx_pos, (unsigned int) rx_anzahl);
							debug_collect(debug_msg);
							// -----

							++rx_pos;
						}
					}// if(rx_anzahl!=0)

					// Checksum ok ??
					if(((unsigned char) hex2asc((checksum >> 4) & 0x0f) == rx_data[rx_pos]) && \
					   ((unsigned char) hex2asc((checksum >> 0) & 0x0f) == rx_data[rx_pos+1]))
					{
						ok=1;
					}
					else
					{
						rx_pos+=2;
						
						// DEBUG
						debug_set_function_name("DapiSeriellFTDISendCommand");
						sprintf(debug_msg, "Daten Empfang: checksum nicht ok !  calc_checksum = %x , rx_checksum = %c %c 0", checksum, rx_data[rx_pos-2], rx_data[rx_pos-1]);
						debug_collect(debug_msg);
						// -----
						
						ok=0;
					}
				}
			}
		}//if(rx_count!=0)
		if(ok!=1)
		{
			// Nicht ok
			--anz_retry;
		}

	} while((ok!=1) && (anz_retry!=0));


	// DEBUG
	debug_set_function_name("DapiSeriellFTDISendCommand");
	sprintf(debug_msg, "retry = %d -----------\r0", (unsigned int) anz_retry);
	debug_collect(debug_msg);
	// -----

	if(ok==1) 
	{
		// DEBUG
		debug_set_function_name("DapiSeriellFTDISendCommand");
		sprintf(debug_msg, "DapiSeriellFTDISendCommand END");
		debug_collect(debug_msg);
		// -----
	
		return DAPI_ERR_NONE;
	}
	else
	{
		// DEBUG
		debug_set_function_name("DapiSeriellFTDISendCommand");
		debug_collect_printall();
		// -----

		return DAPI_ERR_DEV_PKT_OTHER_ERROR;
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiSeriellFTDIWriteByte(unsigned long handle, unsigned long adress, unsigned long value)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='B';		// BYTE
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 4  ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value       ) & 0x0f);

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiSeriellFTDIWriteWord (unsigned long handle,unsigned long adress,unsigned long value)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='W';		// WORD
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 12 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 8  ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 4  ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value       ) & 0x0f);

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiSeriellFTDIWriteLong (unsigned long handle,unsigned long adress,unsigned long value)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='L';		// Long
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 28 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 24 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 20 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 16 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 12 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 8  ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value >> 4  ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((value       ) & 0x0f);

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiSeriellFTDIWriteLongLong (unsigned long handle,unsigned long adress, ULONGLONG value)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned long tx_cnt = 0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='X';		// LongLong
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 60 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 56 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 52 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 48 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 44 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 40 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 36 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 32 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 28 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 24 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 20 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 16 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 12 ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 8  ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value >> 4  ) & 0x0f));
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((value       ) & 0x0f));

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiSeriellFTDIWrite64BytesMax (unsigned long handle,unsigned long adress, unsigned char *data, ULONG data_length)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned long tx_cnt = 0;
	ULONG cnt;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='M';		// Multiple
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	// Length
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((data_length >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((data_length      ) & 0x0f);

	//Data
	for(cnt=0;cnt!=data_length;++cnt)
	{
		tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((*(data+cnt) >> 4 ) & 0x0f));
		tx_buffer[tx_cnt++]=(unsigned char) hex2asc((unsigned long) ((*(data+cnt)      ) & 0x0f));
	}

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiSeriellFTDIReadByte (unsigned long handle,unsigned long adress)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned char rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='B';		// BYTE
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	rx_cnt=2;		// 1 Byte soll empfangen werden

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code == DAPI_ERR_NONE)
	{
		data =((unsigned long) asc2hex(rx[0])<<4) & 0x00f0;
		data|=((unsigned long) asc2hex(rx[1])   ) & 0x000f;
		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiSeriellFTDIReadWord (unsigned long handle,unsigned long adress)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned char rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='W';		// WORD
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	rx_cnt=4;		// 2 Bytes sollen empfangen werden

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code == DAPI_ERR_NONE)
	{
		data =((unsigned long) asc2hex(rx[0])<<12) & 0xf000;
		data|=((unsigned long) asc2hex(rx[1])<<8 ) & 0x0f00;
		data|=((unsigned long) asc2hex(rx[2])<<4 ) & 0x00f0;
		data|=((unsigned long) asc2hex(rx[3])    ) & 0x000f;
		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}
	
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiSeriellFTDIReadLong(unsigned long handle, unsigned long adress)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned char rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;


	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='L';		// LONG
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	rx_cnt=8;		// 4 Bytes sollen empfangen werden

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code == DAPI_ERR_NONE)
	{
		data =((unsigned long) asc2hex(rx[0])<<28) & 0xf0000000;
		data|=((unsigned long) asc2hex(rx[1])<<24) & 0x0f000000;
		data|=((unsigned long) asc2hex(rx[2])<<20) & 0x00f00000;
		data|=((unsigned long) asc2hex(rx[3])<<16) & 0x000f0000;
		data|=((unsigned long) asc2hex(rx[4])<<12) & 0x0000f000;
		data|=((unsigned long) asc2hex(rx[5])<<8 ) & 0x00000f00;
		data|=((unsigned long) asc2hex(rx[6])<<4 ) & 0x000000f0;
		data|=((unsigned long) asc2hex(rx[7])    ) & 0x0000000f;

		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONGLONG DapiSeriellFTDIReadLongLong(unsigned long handle, unsigned long adress)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned char rx[256];
	ULONGLONG data;
	unsigned long tx_cnt = 0;
	unsigned long rx_cnt;

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='X';		// LONGLONG
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	rx_cnt=16;		// 8 Bytes sollen empfangen werden

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if(delib_error_code == DAPI_ERR_NONE)
	{
		#ifdef __linux__
			data =(((ULONGLONG) asc2hex(rx[0] ))<<60) & 0xf000000000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[1] ))<<56) & 0x0f00000000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[2] ))<<52) & 0x00f0000000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[3] ))<<48) & 0x000f000000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[4] ))<<44) & 0x0000f00000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[5] ))<<40) & 0x00000f0000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[6] ))<<36) & 0x000000f000000000LL;
			data|=(((ULONGLONG) asc2hex(rx[7] ))<<32) & 0x0000000f00000000LL;
			data|=(((ULONGLONG) asc2hex(rx[8] ))<<28) & 0x00000000f0000000LL;
			data|=(((ULONGLONG) asc2hex(rx[9] ))<<24) & 0x000000000f000000LL;
			data|=(((ULONGLONG) asc2hex(rx[10]))<<20) & 0x0000000000f00000LL;
			data|=(((ULONGLONG) asc2hex(rx[11]))<<16) & 0x00000000000f0000LL;
			data|=(((ULONGLONG) asc2hex(rx[12]))<<12) & 0x000000000000f000LL;
			data|=(((ULONGLONG) asc2hex(rx[13]))<<8 ) & 0x0000000000000f00LL;
			data|=(((ULONGLONG) asc2hex(rx[14]))<<4 ) & 0x00000000000000f0LL;
			data|=(((ULONGLONG) asc2hex(rx[15]))    ) & 0x000000000000000fLL;
		#else
			data =(((ULONGLONG) asc2hex(rx[0] ))<<60) & 0xf000000000000000;
			data|=(((ULONGLONG) asc2hex(rx[1] ))<<56) & 0x0f00000000000000;
			data|=(((ULONGLONG) asc2hex(rx[2] ))<<52) & 0x00f0000000000000;
			data|=(((ULONGLONG) asc2hex(rx[3] ))<<48) & 0x000f000000000000;
			data|=(((ULONGLONG) asc2hex(rx[4] ))<<44) & 0x0000f00000000000;
			data|=(((ULONGLONG) asc2hex(rx[5] ))<<40) & 0x00000f0000000000;
			data|=(((ULONGLONG) asc2hex(rx[6] ))<<36) & 0x000000f000000000;
			data|=(((ULONGLONG) asc2hex(rx[7] ))<<32) & 0x0000000f00000000;
			data|=(((ULONGLONG) asc2hex(rx[8] ))<<28) & 0x00000000f0000000;
			data|=(((ULONGLONG) asc2hex(rx[9] ))<<24) & 0x000000000f000000;
			data|=(((ULONGLONG) asc2hex(rx[10]))<<20) & 0x0000000000f00000;
			data|=(((ULONGLONG) asc2hex(rx[11]))<<16) & 0x00000000000f0000;
			data|=(((ULONGLONG) asc2hex(rx[12]))<<12) & 0x000000000000f000;
			data|=(((ULONGLONG) asc2hex(rx[13]))<<8 ) & 0x0000000000000f00;
			data|=(((ULONGLONG) asc2hex(rx[14]))<<4 ) & 0x00000000000000f0;
			data|=(((ULONGLONG) asc2hex(rx[15]))    ) & 0x000000000000000f;
		#endif
		
		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiSeriellFTDIRead64BytesMax (unsigned long handle,unsigned long adress, unsigned char *data, ULONG data_length)
{
	unsigned long delib_error_code;
	unsigned char tx_buffer[256];
	unsigned long tx_cnt = 0;
	ULONG i;
	unsigned char rx[256];
	ULONG rx_cnt;

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='M';		// Multiple
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 12) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 8 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((adress      ) & 0x0f);

	// Length
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((data_length >> 4 ) & 0x0f);
	tx_buffer[tx_cnt++]=(unsigned char) hex2asc((data_length      ) & 0x0f);

	rx_cnt=data_length*2;		// n Bytes sollen empfangen werden

	PacketStatisticStart(handle);
	delib_error_code = DapiSeriellFTDISendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);

	if(delib_error_code == DAPI_ERR_NONE)
	{
		for(i=0;i!=data_length;++i)
		{
			*(data + i) = ((unsigned char) asc2hex(rx[i*2 + 0]) << 4);
			*(data + i)|= ((unsigned char) asc2hex(rx[i*2 + 1])     );
		}
		return;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return;
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
