//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_io_tcp.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2020
//  web: http://www.deditec.de/
//  mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
//#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiOpenModuleIntern_ETH
// DapiCloseModuleIntern_ETH
// ------------------------------------
// tcp_tx_open
// tcp_tx_send_rec_data
// tcp_tx_close
// tcp_rx_close_session
// tcp_tx_send_rec_data_big
// ------------------------------------
// DapiTCPSendCommand
// DapiTCPSendCommandBig
// DapiTCPSendSpecialCommand
// DapiTCPSpecialCmd
// ------------------------------------
// DapiTCPWriteByte
// DapiTCPWriteWord
// DapiTCPWriteLong
// DapiTCPWriteLongLong
// ------------------------------------
// DapiTCPReadByte
// DapiTCPReadWord
// DapiTCPReadLong
// DapiTCPReadLongLong
// DapiTCPWriteMultipleBytes
// DapiTCPReadMultipleBytes
// ------------------------------------
// DapiTCPReadFifo

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include <sys/socket.h>	// send()
#include <unistd.h> // read(), close()


#include "proj_config.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------

	#include "library/vc/deditec_buff_utils.h"
	#include "library/vc/deditec_tcp_io.h"
	#include "library/delib/delib.h"
	#include "library/delib/delib_intern.h"
	#include "library/communication/dt_communication_paramams.h"
	#include "library/embedded/deditec_debug_comfort.h"

	#include "delib_packet_statistic.h"
	#include "delib_hdlmgr.h"
	#include "delib_set_get_error.h"
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include "library/vc/comport.h"
	#include "library/ftdi/ftd2xx.h"
	#include "library/delib/delib.h"
	#include "library/delib/delib_intern.h"
	#include "library/delib/delib_extended_functions.h"
	#include "delib_set_get_error.h"
	#include "library/vc/deditec_buff_utils.h"
	#include "library/vc/deditec_string_utils.h"
	#include "library/vc/deditec_tcp_io.h"
	#include "library/embedded/dt_embedded_parameter.h"
	#include "delib_open_close.h"
#endif

// ------------------------------------
// ------------------------------------
// Defines
// ------------------------------------
// ------------------------------------

#define BUFF_TCP_RECV_SIZE 				16384  	// better don't touch, used by read() and recv()
#define TX_BUFFER_CP_LENGTH 			1200

#ifdef ENABLE_TCP_ENCRYPTION
#define TCP_RO_TX_ENCRYPTION_START		4		// id(2) + ro-cmd(1) + job_id(1)
#define TCP_RO_RX_ENCRYPTION_START		5		// id(2) + answer(1) + ok(1) + job_id(1)
#endif

#define TCP_RO_1_TX_HEADER_LENGTH 		6		// id(2) + ro-cmd(1) + job_id(1) + length(2)
#define TCP_RO_1_RX_HEADER_LENGTH 		7		// id(2) + answer(1) + ok(1) + job_id(1) + length(2)
#define TCP_RO_BIG_TX_HEADER_LENGTH 	8		// id(2) + ro-cmd(1) + job_id(1) + length(4)
#define TCP_RO_BIG_RX_HEADER_LENGTH 	9		// id(2) + answer(1) + ok(1) + job_id(1) + length(4)

// ------------------------------------
// ------------------------------------
// Declarations
// ------------------------------------
// ------------------------------------

#ifdef ENABLE_TCP_ENCRYPTION
	extern unsigned long dt_encryption_get_info(	DAPI_HANDLE *DapiHandle,
													char *encryption_pw, unsigned long encryption_pw_size,
													unsigned char *packet_id, bool *encryption_needed);
													
	extern unsigned long dt_encrypt_tcp_pre(DAPI_HANDLE *DapiHandle, unsigned char *tcp_msg, unsigned long *tcp_msg_len, unsigned long *tcp_msg_pos);
	extern unsigned long dt_encrypt_tcp_msg(DAPI_HANDLE *DapiHandle, unsigned char *tcp_msg, unsigned long *tcp_msg_len);

	extern unsigned long dt_decrypt_tcp_pre(DAPI_HANDLE *DapiHandle, unsigned long *tcp_msg_size, unsigned long *tcp_msg_pos);
	extern unsigned long dt_decrypt_tcp_msg(DAPI_HANDLE *DapiHandle, unsigned char *tcp_msg, unsigned long tcp_msg_len);
#endif

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

#ifdef __linux__
	#ifdef DEBUG_OUTPUT
		#define debug_print(x) printf("%s\n", x)
	#endif
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_open(ULONG handle32, unsigned char * host_name, unsigned long host_name_size, unsigned long port_no, unsigned long tcp_timeout)
{
	char debug_msg[200];
	unsigned long ret;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	TCP_IO * tcp_params; // creates global pointer, pointing to a structure of type TCP_IO (see deditec_tcp_io.h)
	tcp_params = (TCP_IO *) malloc(sizeof(TCP_IO)); // allocate ram space for struct TCP_IO beginning at pointer

	if (tcp_params == 0)
	{
		// DEBUG	
		debug_set_function_name("tcp_tx_open");
		sprintf(debug_msg,"return value is %d dec ", -5);
		debug_error(debug_msg);
		// -----

		return -5; // error allocating memory, failed
	}

	if(buffcpy_n(tcp_params->host_name, IP_ADDR_SERVER_SIZE, (unsigned char *) host_name, host_name_size) < 0)
	{
		free(tcp_params);

		return -6; // dest_buffer too small
	}
	
	tcp_params->host_name_size = host_name_size;
	tcp_params->port_no = port_no;

	// DEBUG
	debug_set_function_name("tcp_tx_open");
	sprintf(debug_msg,"tcp_params = 0x%lx", (unsigned long) tcp_params);//Sonst gehts unter Linux nicht ....
	debug_info(debug_msg);
	sprintf(debug_msg,"using IP: %s", tcp_params->host_name);
	debug_info(debug_msg);
	sprintf(debug_msg,"using port: %d", (unsigned int) tcp_params->port_no);
	debug_info(debug_msg);
	sprintf(debug_msg,"using timeout: %d", (unsigned int) tcp_timeout);
	debug_info(debug_msg);
	// -----

	ret = tcp_client_connect(tcp_params, host_name, host_name_size, tcp_params->port_no, 1, tcp_timeout); // returns pointer to tcp_params or neg. error code

	if (ret==-1 || ret==-2 || ret==-3 || ret==-4 || ret==-5|| ret==-6) // socket error
	{
		// DEBUG
		debug_set_function_name("tcp_tx_open");
		sprintf(debug_msg, "socket error - return is %d dec", (unsigned int) ret);
		debug_error(debug_msg);
		// -----

		free(tcp_params);

		return ret;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_open");
	sprintf(debug_msg,"tcp_params->tcp_handle_tx = %d dez", (unsigned int) tcp_params->tcp_handle_tx);
	debug_info(debug_msg);
	sprintf(debug_msg,"function end - return value: 0x%x", (unsigned int) ret);
	debug_info(debug_msg);
	// -----

//	DapiHandle->TCPHandle = tcp_params;
	DapiHandle->tcp_params = tcp_params;
	DapiHandle->TCPHandle_tx = tcp_params->tcp_handle_tx;
	return RETURN_OK;//(unsigned long) tcp_params; // returns pointer to tcp_params
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_send_rec_data( ULONG handle32, unsigned long tcp_handle_tx, unsigned long job_id, unsigned long tcp_timeout,
									unsigned long command_type,
									unsigned char * tx_buffer, unsigned long tx_buffer_size,
									unsigned char * rx_buffer, unsigned long rx_buffer_size, unsigned long *rx_buffer_length)
{
	unsigned char tx_buffer_cp[TX_BUFFER_CP_LENGTH];
	unsigned char *recv_buffer;
	unsigned long recv_buffer_size;
	unsigned long recv_error;
	unsigned long recv_pos;
	unsigned long recv_len = 0;
	signed long amount_bytes_sent;
	signed   long amount_bytes_received;
	unsigned long tx_buffer_cnt;
	unsigned long ret;
	struct timeval time_tcp_timeout;
	unsigned long ret_select_bind;
	fd_set set;
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	
	unsigned long err;
	unsigned char dt_packet_id_1;
	
	unsigned long tx_data_pos;
	unsigned long rx_data_pos;
	
	unsigned long i;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	// DEBUG
	debug_collect_init();
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "tcp_tx_send_rec_data START");
	debug_collect(debug_msg);
	sprintf(debug_msg, "tcp_tx_send_rec_data(handle=0x%x, tx_buff_len=%ddez, rx_buff_len=%ddez)", (unsigned int) tcp_handle_tx, (unsigned int) tx_buffer_size, (unsigned int) rx_buffer_size);
	debug_collect(debug_msg);
	// -----
	
	// timeout setzen
	time_tcp_timeout.tv_sec = (tcp_timeout / 1000);
	time_tcp_timeout.tv_usec = (tcp_timeout % 1000) * 1000;
		
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "tcp_timeout has been set to %ld sec and %ld usec", time_tcp_timeout.tv_sec, time_tcp_timeout.tv_usec);
	debug_collect(debug_msg);
	// -----
		
	if (tcp_handle_tx < 0) 
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "tcp_handle_tx < 0");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		return DAPI_ERR_GEN_SOCKET_ERROR;
	}
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "tcp_handle_tx = 0x%x", (unsigned int) tcp_handle_tx);
	debug_collect(debug_msg);
	// -----
	
	#ifdef ENABLE_TCP_ENCRYPTION
		if (dt_encryption_get_info(DapiHandle, NULL, 0, &dt_packet_id_1, NULL) != DAPI_ERR_NONE)
		{
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data");
			sprintf(debug_msg, "unknown encryption_type - DapiHandle->encryption_type=%ld - return=0x%lx", (unsigned long) DapiHandle->encryption_type, (unsigned long) DAPI_ERR_GEN_UNKNOWN_ENCRYPTION_TYPE);
			debug_collect(debug_msg);
			debug_collect_printall();
			// -----
			
			return DAPI_ERR_GEN_UNKNOWN_ENCRYPTION_TYPE;
		}
	#else
		dt_packet_id_1 = (unsigned char) DEDITEC_TCP_PACKET_ID_1_NORMAL;
	#endif
	
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// SENDING MESSAGE
	
	// ---- creating message header (3 bytes for ID's, 1 byte for job_id, 2 bytes for message length-> 6 bytes total)
	tx_buffer_cnt = 0;

	// now 3 bytes for DEDITEC ID and the command-id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_0;
	tx_buffer_cp[tx_buffer_cnt++] = dt_packet_id_1;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) command_type;

	// now first byte for job_id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) job_id;
	
	tx_data_pos = TCP_RO_1_TX_HEADER_LENGTH;
	
	
	#ifdef ENABLE_TCP_ENCRYPTION
		if (dt_encrypt_tcp_pre(DapiHandle, tx_buffer_cp, &tx_buffer_cnt, &tx_data_pos) != DAPI_ERR_NONE)
		{
			// DEBUG
			debug_set_function_name("dt_encrypt_tcp_pre");
			sprintf(debug_msg, "buffer too small return is %d dec", DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR);
			debug_collect(debug_msg);
			debug_collect_printall();
			// -----

			return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;		
		}
	#endif
	
	// now 2 bytes for length of receive buffer
	tx_buffer_cp[tx_buffer_cnt++] = ((unsigned char) ((tx_buffer_size + tx_data_pos) >> 8) & 255); // rotate by 8 bits and mask upper 8 bits
	tx_buffer_cp[tx_buffer_cnt++] =  (unsigned char) ((tx_buffer_size + tx_data_pos) & 255);       // mask lower 8 bits

	if (buffcpy_n(	(tx_buffer_cp + tx_data_pos), (TX_BUFFER_CP_LENGTH - tx_data_pos), tx_buffer, tx_buffer_size) != RETURN_OK)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");
		sprintf(debug_msg, "buffer too small return is %d dec", DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----

		return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;
	}

	tx_buffer_cnt += tx_buffer_size;
	
	#ifdef ENABLE_TCP_ENCRYPTION
		if (dt_encrypt_tcp_msg(DapiHandle, tx_buffer_cp, &tx_buffer_cnt) != DAPI_ERR_NONE)
		{
			debug_set_function_name("dt_encrypt_tcp_msg");
			sprintf(debug_msg, "Error while encrypting tx_buffer - return=%d", DAPI_ERR_GEN_ENCRYPTION_ERROR);
			debug_collect(debug_msg);
			debug_collect_printall();
			
			return DAPI_ERR_GEN_ENCRYPTION_ERROR;		
		}
	#endif
	
	amount_bytes_sent = send(tcp_handle_tx, (const char *) tx_buffer_cp, tx_buffer_cnt, 0);
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "SENDING %ld bytes: / send(...) returns %ld", tx_buffer_cnt, amount_bytes_sent);
	debug_collect(debug_msg);
	if (amount_bytes_sent > 0)
	{
		for (i=0; i < (unsigned long) amount_bytes_sent; i+=16)
		{
			debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), tx_buffer_cp, sizeof(tx_buffer_cp), i, amount_bytes_sent);
			debug_collect(debug_msg);
		}
	}
	// -----

	if(amount_bytes_sent < 0)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");
		sprintf(debug_msg, "SEND ERROR: return=%d", DAPI_ERR_COM_CONN_COULD_NOT_BE_ESTABLISHED);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----

		return DAPI_ERR_COM_CONN_COULD_NOT_BE_ESTABLISHED;
	}
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// RECEIVING MESSAGE
	
	rx_data_pos = TCP_RO_1_RX_HEADER_LENGTH;
	recv_buffer_size = TCP_RO_1_RX_HEADER_LENGTH + rx_buffer_size;
	
	#ifdef ENABLE_TCP_ENCRYPTION
		if ((ret = dt_decrypt_tcp_pre(DapiHandle, &recv_buffer_size, &rx_data_pos)) != DAPI_ERR_NONE)
		{
			debug_set_function_name("dt_decrypt_tcp_pre");
			sprintf(debug_msg, "Error while calculating rx_buffer_size - return=%ld", (unsigned long) ret);
			debug_collect(debug_msg);
			debug_collect_printall();
			
			return ret;		
		}
	#endif
	
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "malloc(recv_buffer_size=%ld) now....", recv_buffer_size);
	debug_collect(debug_msg);

	recv_buffer = (unsigned char *) malloc(recv_buffer_size);
	if (recv_buffer == NULL)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");
		sprintf(debug_msg, "malloc() recv_buffer failed - return=%d", DAPI_ERR_GEN_MALLOC_ERROR);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		return DAPI_ERR_GEN_MALLOC_ERROR;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "%lu bytes memory allocated for recv_buffer", recv_buffer_size);
	debug_collect(debug_msg);
	sprintf(debug_msg, "setting tcp_timeout * tcp_timeout=%ld", tcp_timeout);
	debug_collect(debug_msg);
	// -----
	
	FD_ZERO(&set);

	// Den Socket, der Verbindungen annimmt hinzufgen
	FD_SET(tcp_handle_tx, &set);
	
	ret_select_bind = select(tcp_handle_tx + 1 , &set, NULL, NULL, &time_tcp_timeout);
	amount_bytes_received = -1;

	if (FD_ISSET(tcp_handle_tx, &set)) 
	{
		#ifdef __linux__
			amount_bytes_received = read(tcp_handle_tx, recv_buffer, recv_buffer_size);
		#else
			amount_bytes_received = recv(tcp_handle_tx, (char *) recv_buffer , recv_buffer_size, 0);
		#endif
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");

	sprintf(debug_msg, "RECEIVING %ld bytes", amount_bytes_received);
	debug_collect(debug_msg);
	if(amount_bytes_received > 0)
	{
		for (i=0; i < (unsigned long) amount_bytes_received; i+=16)
		{
			debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), recv_buffer, recv_buffer_size, i, amount_bytes_received);
			debug_collect(debug_msg);
		}/*xx*/
	}
	// -----

	if (amount_bytes_received < 0) 
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "recv return < 0 (ret-value = %d) - function end - return is DAPI_ERR_COM_DEVICE_DID_NOT_ANSWER", (unsigned int) amount_bytes_received);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----

		free(recv_buffer);
		return DAPI_ERR_COM_DEVICE_DID_NOT_ANSWER; // failed to receive message from server
	}
	
	#ifdef ENABLE_TCP_ENCRYPTION
		if ((ret = dt_decrypt_tcp_msg(DapiHandle, recv_buffer, amount_bytes_received)) != DAPI_ERR_NONE)
		{
			debug_set_function_name("dt_decrypt_tcp_msg");
			sprintf(debug_msg, "Error while decrypting recv_buffer - return=%ld", (unsigned long) ret);
			debug_collect(debug_msg);
			debug_collect_printall();
		
			free(recv_buffer);		
			return ret;		
		}
	#endif
	
	// antwort berprfen
	recv_error = 0;
	recv_pos = 0;

	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_0)					++recv_error;
	if (recv_buffer[recv_pos++] != dt_packet_id_1)							++recv_error;
	if (recv_buffer[recv_pos++] != (0x80 | command_type))					++recv_error;
	recv_pos++;
	if (recv_buffer[recv_pos++] != (job_id & 0xff))							++recv_error;

	if (recv_error != 0)	// header not ok = error
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "error while receiving message: header of rx_msg not ok! - No of recv_errors=%d", (unsigned int) recv_error);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----

		free(recv_buffer);
		return DAPI_ERR_DEV_PACKET_HEADER_NOT_OK;
	}

	// device error code
	if (recv_buffer[3] != 0x0)
	{
		err = (recv_buffer[3] & 0xff);
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "error while receiving message: recv_buffer[3] != 0x0  ->  recv_buffer[3] = 0x%lx", err);
		debug_collect(debug_msg);

		// Hier haben wir kein Protokollproblem, deshalb immer debuggen ... zur Info 
		debug_print(debug_msg);
		// -----

		free(recv_buffer);
		return (err | DAPI_ERR_DEV_CLASS);
	}

	if (amount_bytes_received > 0)  // seems possible to receive an empty tcp-packet
	{
		ret = buffcpy_n(rx_buffer, rx_buffer_size, recv_buffer+rx_data_pos, rx_buffer_size); // -1 if rx_buffer too small

		if (ret < 0)
		{		
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----
			
			free(recv_buffer);
			return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;
		}
	}

	if(rx_buffer_length != NULL)
	{
		recv_len = (*(recv_buffer + rx_data_pos - 2) << 8);
		recv_len |= *(recv_buffer + rx_data_pos - 1);
		recv_len -= rx_data_pos;

		*rx_buffer_length = recv_len;
	}

	free(recv_buffer);

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");

	sprintf(debug_msg, "tcp_tx_send_rec_data END * return=DAPI_ERR_NONE");
	debug_collect(debug_msg);
	// -----

	return DAPI_ERR_NONE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_close(TCP_IO* tcp_params_)
{
//	char debug_msg[200];
	unsigned int ret;
	ret = tcp_disconnect(tcp_params_);
	free(tcp_params_);

	// DEBUG
//	debug_set_function_name("tcp_tx_close");
//	sprintf(debug_msg,"function end - return is %d dec", ret);
//	debug_info(debug_msg);
	// -----

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifdef __linux__
   unsigned long tcp_rx_close_session(unsigned long socket_file_descr) //closing server session
   {
      return close(socket_file_descr);
   }
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_send_rec_data_big(	ULONG handle32, unsigned long   tcp_handle_tx, unsigned long job_id, unsigned long   tcp_timeout,
										unsigned long command_type,
										unsigned char * tx_buffer, unsigned long tx_buffer_size,
										unsigned char * rx_buffer, unsigned long rx_buffer_size, unsigned long *rx_buffer_length)
{
	unsigned char *tx_buffer_cp;
	unsigned long tx_buffer_cp_size;
	unsigned char *recv_buffer;
	unsigned long recv_buffer_size;
	unsigned long recv_error;
	unsigned long recv_pos;
	unsigned long amount_bytes_sent;
	signed   long amount_bytes_received;
	unsigned long tx_buffer_cnt;
	unsigned long ret;
	struct timeval time_tcp_timeout;
	unsigned long ret_select_bind;
	fd_set set;
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	unsigned long i;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	// DEBUG
	debug_collect_init();
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "tcp_tx_send_rec_data_big START");
	debug_collect(debug_msg);
	sprintf(debug_msg, "tcp_tx_send_rec_data_big(handle=0x%x, tx_buff_len=%ddez, rx_buff_len=%ddez)", (unsigned int) tcp_handle_tx, (unsigned int) tx_buffer_size, (unsigned int) rx_buffer_size);
	debug_collect(debug_msg);
	// -----
	
	tx_buffer_cp_size = TCP_RO_BIG_TX_HEADER_LENGTH + tx_buffer_size;
	tx_buffer_cp = (unsigned char *) malloc(tx_buffer_cp_size);

	if (tx_buffer_cp == NULL)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "malloc() tx_buffer_cp failed");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		return DAPI_ERR_GEN_MALLOC_ERROR;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "%lu bytes memory allocated for tx_buffer_cp", tx_buffer_cp_size);
	debug_collect(debug_msg);
	// -----
	
	recv_buffer_size = TCP_RO_BIG_RX_HEADER_LENGTH + rx_buffer_size;
	recv_buffer = (unsigned char *) malloc(recv_buffer_size);

	if (recv_buffer == NULL)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "malloc() recv_buffer failed");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		free(tx_buffer_cp);	
		return DAPI_ERR_GEN_MALLOC_ERROR;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "%lu bytes memory allocated for recv_buffer", recv_buffer_size);
	debug_collect(debug_msg);
	// -----
	
	#ifndef __linux__
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	
		// timeout wird nur bei windows gesetzt
		time_tcp_timeout.tv_sec = (tcp_timeout / 1000);
		time_tcp_timeout.tv_usec = (tcp_timeout % 1000) * 1000;
		
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "tcp_timeout has been set to %ld sec and %ld usec", time_tcp_timeout.tv_sec, time_tcp_timeout.tv_usec);
		debug_collect(debug_msg);
		// -----
		
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	#endif
	
	if (tcp_handle_tx < 0)
	{	
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "tcp_handle_tx < 0");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		free(tx_buffer_cp);
		free(recv_buffer);
		return DAPI_ERR_GEN_SOCKET_ERROR; // return socket error
	}
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "tcp_handle_tx = 0x%x", (unsigned int) tcp_handle_tx);
	debug_collect(debug_msg);
	// -----
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// SENDING MESSAGE
	
	
	// ---- creating message header (3 bytes for ID's, 1 byte for job_id, 4 bytes for message length-> 8 bytes total)
	tx_buffer_cnt = 0;

	// now 3 bytes for DEDITEC ID and the command-id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_0;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_1_NORMAL;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_2_CMD_RO_BIG;

	// now first byte for job_id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) (job_id & 0xff);

	// now 4 bytes for length of receive buffer
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size >> 24) & 255);
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size >> 16) & 255);
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size >> 8 ) & 255);
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size      ) & 255);

	ret = buffcpy_n( tx_buffer_cp + tx_buffer_cnt,  tx_buffer_cp_size - tx_buffer_cnt, 
					 tx_buffer, 					tx_buffer_size);

	if (ret != 0)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "buffer too small return is %d dec", -2);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----

		free(tx_buffer_cp);
		free(recv_buffer);		
		return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;
	}

	tx_buffer_cnt += tx_buffer_size;
	amount_bytes_sent = send(tcp_handle_tx, (const char *) tx_buffer_cp, tx_buffer_cnt, 0);

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "SENDING %ld bytes:", amount_bytes_sent);
	debug_collect(debug_msg);
	if (amount_bytes_sent > 0)
	{
		for (i=0; i<amount_bytes_sent; i+=16)
		{
			debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), tx_buffer_cp, sizeof(tx_buffer_cp), i, amount_bytes_sent);
			debug_collect(debug_msg);
		}
	}
	// -----
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// RECEIVING MESSAGE
	
	
	amount_bytes_received = 0;

	#ifdef __linux__
		amount_bytes_received = read(tcp_handle_tx, recv_buffer, recv_buffer_size);
	#else
		FD_ZERO(&set);

		// Den Socket, der Verbindungen annimmt hinzufgen
		FD_SET(tcp_handle_tx, &set);
		
		ret_select_bind = select(tcp_handle_tx , &set, NULL, NULL, &time_tcp_timeout);
		amount_bytes_received=-1;

		if (FD_ISSET(tcp_handle_tx, &set)) 
		{
			amount_bytes_received = recv(tcp_handle_tx, (char *) recv_buffer , recv_buffer_size, 0);
		}
	#endif

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");

	sprintf(debug_msg, "RECEIVING %ld bytes", amount_bytes_received);
	debug_collect(debug_msg);
	if (amount_bytes_received > 0)
	{
		for (i=0; i < (unsigned long) amount_bytes_received; i+=16)
		{
			debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), recv_buffer, recv_buffer_size, i, amount_bytes_received);
			debug_collect(debug_msg);
		}
	}
	// -----
	
	// antwort berprfen
	recv_error = 0;
	recv_pos = 0;

	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_0)					++recv_error;
	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_1_NORMAL)			++recv_error;
	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_2_ANSWER_RO_BIG)	++recv_error;
	if (recv_buffer[recv_pos++] != 0x0)										++recv_error;
	if (recv_buffer[recv_pos++] != (job_id & 0xff))							++recv_error;

	if (recv_error != 0)	// error
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");

		sprintf(debug_msg, "error while receiving message: header of rx_msg not ok! - recv_error=%d", (unsigned int) recv_error);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		free(tx_buffer_cp);
		free(recv_buffer);
		return DAPI_ERR_DEV_PACKET_HEADER_NOT_OK; // failed to receive message from server
	}
	else	// header ok
	{
		if(buffcpy_n(rx_buffer, rx_buffer_size, recv_buffer+TCP_RO_BIG_RX_HEADER_LENGTH, amount_bytes_received-TCP_RO_BIG_RX_HEADER_LENGTH) == -1) // -1 if rx_buffer too small
		{
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data_big");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", -5);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----
			
			free(recv_buffer);
			free(tx_buffer_cp);
			return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;
		}
	}

	if (amount_bytes_received < 0) 
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");

		sprintf(debug_msg, "recv return <0 (ret-value = %d) - function end - return is -3", (unsigned int) amount_bytes_received);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----

		free(tx_buffer_cp);
		free(recv_buffer);
		
		return DAPI_ERR_COM_DEVICE_DID_NOT_ANSWER; // failed to receive message from server
	}

	if (amount_bytes_received > 0)  // seems possible to receive an empty tcp-packet
	{		
		ret = buffcpy_n(rx_buffer, rx_buffer_size, recv_buffer+TCP_RO_BIG_RX_HEADER_LENGTH, amount_bytes_received-TCP_RO_BIG_RX_HEADER_LENGTH); // -1 if rx_buffer too small

		if (ret < 0)
		{
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data_big");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", -5);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----

			free(tx_buffer_cp);
			free(recv_buffer);
			return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;
		}
	}

	free(tx_buffer_cp);
	free(recv_buffer);
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");

	sprintf(debug_msg, "tcp_tx_send_rec_data_big END * return=%ld", (amount_bytes_received - TCP_RO_1_RX_HEADER_LENGTH) );
	debug_collect(debug_msg);
	// -----
	
	if(rx_buffer_length != NULL)
	{
		*rx_buffer_length = amount_bytes_received - TCP_RO_BIG_RX_HEADER_LENGTH;
	}

	return DAPI_ERR_NONE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiOpenModuleIntern_ETH(ULONG handle32, ULONG moduleID, ULONG moduleNr_, unsigned char* exbuffer)
{
	char debug_msg[200];
	unsigned char ip_addr[IP_ADDR_SERVER_SIZE];
	ULONG ret;
	char reg_key_name[60];
	char reg_key_param[70];
	ULONG reg_key_value_port;
	ULONG reg_key_value_enabled;
	ULONG reg_key_value_timeout;
	ULONG reg_key_value_encryption_type;
	unsigned char reg_key_encryption_password[32];
	unsigned char temp[33];
	ULONG i;
	ULONG len;
	ULONG err;
	ULONG moduleNr = moduleNr_ & MAX_NO_OF_MODULES_MASK;
	//unsigned long delib_error_code;

	DAPI_OPENMODULEEX_STRUCT* exbuffer_int;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	if(exbuffer == 0)
	{
		#ifndef __linux__
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// --------------------------------------------------------------------
			DapiInternGetRegistryKeyName(moduleID, (char*) reg_key_name, sizeof(reg_key_name));
			// --------------------------------------------------------------------
			//
			// ------------------------------------------------
			// Registry read Server enabled
			sprintf(reg_key_param, "Enabled%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_enabled);
			if (err == RETURN_OK)
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_enabled);
					debug_info(debug_msg);
				#endif
				// -----
			}
			else
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Port-No
			sprintf(reg_key_param, "Port%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_port);
			if (err == RETURN_OK)
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_port);
					debug_info(debug_msg);
				#endif
				// -----
			}
			else
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Timeout
			sprintf(reg_key_param, "Timeout%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_timeout);
			if (err == RETURN_OK)
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_timeout);
					debug_info(debug_msg);
				#endif
				// -----
			}
			else
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----

				// default value
				reg_key_value_timeout = 5000;

				debug_error("Parameter not found in Registry: Using default Value: timeout=5000");
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Hostname
			sprintf(reg_key_param, "Hostname%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_STRING, reg_key_name, reg_key_param, (char*) ip_addr, IP_ADDR_SERVER_SIZE, (ULONG*) 0);
			if (err == RETURN_OK)
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_STRING: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=\"%s\"\n", reg_key_name, reg_key_param, ip_addr);
					debug_info(debug_msg);
				#endif
				// -----
			}
			else
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_STRING: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Hostname
			sprintf(reg_key_param, "EncryptionType%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_encryption_type);
			if (err == RETURN_OK)
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_encryption_type);
					debug_info(debug_msg);
				#endif
				// -----
			}
			else
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----

				// default value
				reg_key_value_encryption_type = DAPI_OPEN_MODULE_ENCRYPTION_TYPE_NONE;

				debug_error("Parameter not found in Registry: Using default Value: timeout=5000");
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Hostname
			sprintf(reg_key_param, "EncryptionPassword%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_STRING, reg_key_name, reg_key_param, (char*) temp, sizeof(temp), (ULONG*) 0);
			if (err == RETURN_OK)
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_STRING: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=OK\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----

				len = strlen_n((char*)temp, sizeof(temp));

				if ((len % 2) == 0)
				{
					len = len / 2;

					for (i=0; i!=len; i++)
					{
						reg_key_encryption_password[i]  =  (unsigned char) (asc2hex(temp[i*2+0]) << 4);
						reg_key_encryption_password[i] |=  (unsigned char) (asc2hex(temp[i*2+1]) << 0);
					}

					reg_key_encryption_password[len] = 0;

					if (dt_decrypt_buffer(	DEDITEC_ENCRYPTION_TEMPORARY_ADMIN_PW, strlen_n(DEDITEC_ENCRYPTION_TEMPORARY_ADMIN_PW, sizeof(DEDITEC_ENCRYPTION_TEMPORARY_ADMIN_PW)),
											reg_key_encryption_password, strlen_n((char*) reg_key_encryption_password, sizeof(reg_key_encryption_password))) != RETURN_OK)
					{
						debug_error("Error while decrypting EncryptionPassword\n");
					}
				}
				else
				{
					debug_error("Error EncryptionPassword has an invalid length\n");
				}
			}
			else
			{
				// DEBUG
				#ifdef DEBUG_OUTPUT
					sprintf(debug_msg,"GET_REGISTRY_STRING: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
					debug_info(debug_msg);
				#endif
				// -----
			}
			// ------------------------------------------------
			
		#endif
	}
	else
	{
		// exbuffer wird benutzt
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"exbuffer read");
		debug_info(debug_msg);
		// -----

		exbuffer_int = (DAPI_OPENMODULEEX_STRUCT*) exbuffer;

		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"ip=%s", exbuffer_int->address);											debug_info(debug_msg);
		sprintf(debug_msg,"port=%d", (unsigned int) exbuffer_int->portno);							debug_info(debug_msg);
		sprintf(debug_msg,"timeout=%ld", (unsigned long) exbuffer_int->timeout);					debug_info(debug_msg);
		sprintf(debug_msg,"encryption_type=%ld", (unsigned long) exbuffer_int->encryption_type);	debug_info(debug_msg);
		// -----

		reg_key_value_enabled 			= 1;
		reg_key_value_port 				= exbuffer_int->portno;
		reg_key_value_timeout 			= exbuffer_int->timeout;
		reg_key_value_encryption_type 	= exbuffer_int->encryption_type;
		strcpy_n((char*) reg_key_encryption_password, sizeof(reg_key_encryption_password), (char*) exbuffer_int->encryption_password, sizeof(exbuffer_int->encryption_password));
		strcpy_n((char*) ip_addr, sizeof(ip_addr), (char*) exbuffer_int->address, sizeof(exbuffer_int->address));
	}

	if((moduleNr_ & DAPI_OPEN_MODULE_OPTION_DISABLE_ENCRYPTION) != 0)
	{
		reg_key_value_encryption_type = DAPI_OPEN_MODULE_ENCRYPTION_TYPE_NONE;

		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"moduleNr_ & DAPI_OPEN_MODULE_OPTION_DISABLE_ENCRYPTION != 0 -> DAPI_OPEN_MODULE_ENCRYPTION_TYPE_NONE");;
		debug_info(debug_msg);
		// -----
	}

	if(reg_key_value_enabled != 1)
	{
		// Enabled in registry is not =1
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"enabled !=1 -> return 0");;
		debug_info(debug_msg);
		// -----

		return DAPI_ERR_COM_CONN_COULD_NOT_BE_ESTABLISHED;
	}
	else
	{
		if(reg_key_value_port==0)
		{
			// Default Port
			reg_key_value_port = DEDITEC_DEFAULT_TCP_PORT_NO;
		}
		
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"trying to connect to server %s:%d", ip_addr, (unsigned int) reg_key_value_port);
		debug_info(debug_msg);
		// -----

		ret = tcp_tx_open(handle32, ip_addr, sizeof(ip_addr), reg_key_value_port, reg_key_value_timeout); //handle is false, it contains adr. of tcp_params

		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"tcp_tx_open returns %d dez", (unsigned int) ret);
		debug_info(debug_msg);
		// -----

		if (ret==-1 || ret==-2 || ret==-3 || ret==-4 || ret==-5|| ret==-6) // socket error
		{
			// Connect Error
			DapiHandle->tcp_params = NULL;
			return DAPI_ERR_COM_DEVICE_DID_NOT_ANSWER;
		}
		else
		{
			DapiHandle->SubDevice = 2;
			DapiHandle->timeout = reg_key_value_timeout;

			DapiHandle->encryption_type = reg_key_value_encryption_type;
			strcpy_n((char*) DAPI_PTR_SIZE DapiHandle->encryption_password, sizeof(DapiHandle->encryption_password), (char*) reg_key_encryption_password, sizeof(reg_key_encryption_password));
			strcpy_n((char*) DAPI_PTR_SIZE DapiHandle->ip_addr, sizeof(DapiHandle->ip_addr), (char*) ip_addr, sizeof(ip_addr));
	
			sprintf(debug_msg,"delib_io_tcp.cpp IP-adr: DapiHandle->ip_addr = %s", DapiHandle->ip_addr);
			debug_info(debug_msg);
		}
	}
	
	return DAPI_ERR_NONE;	// -> ok
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPSendCommand(ULONG handle32, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl)
{
	//unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiTCPSendCommand");
		sprintf(debug_msg,"tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, tx[0],tx[1], (unsigned int) rx_anzahl);
		debug_print(debug_msg);
	#endif
	// -----

	++DapiHandle->jobID;
	
	return tcp_tx_send_rec_data(handle32, DapiHandle->TCPHandle_tx, DapiHandle->jobID, DapiHandle->timeout, DEDITEC_TCP_PACKET_ID_2_CMD_RO_1, tx, tx_length, rx, rx_anzahl, NULL);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPSendCommandBig(ULONG handle32, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl, unsigned long *rx_data_length)
{
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiTCPSendCommandBig");
		sprintf(debug_msg, "tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, tx[0],tx[1], (unsigned int) rx_anzahl);
		debug_print(debug_msg);
	#endif
	// -----

	++DapiHandle->jobID;
		
	if ((delib_error_code = tcp_tx_send_rec_data_big(handle32, DapiHandle->TCPHandle_tx, DapiHandle->jobID, DapiHandle->timeout, DEDITEC_TCP_PACKET_ID_2_CMD_RO_1, tx, tx_length, rx, rx_anzahl, rx_data_length)) != DAPI_ERR_NONE)
	{
		*rx_data_length = 0;
	}
	
	return delib_error_code;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPSendSpecialCommand(ULONG handle32, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl, unsigned long *rx_cnt)
{
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		char debug_msg[200];
		debug_set_function_name("DapiTCPSendSpecialCommand");
		sprintf(debug_msg, "tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, tx[0],tx[1], (unsigned int) rx_anzahl);
		debug_print(debug_msg);
	#endif
	// -----

	++DapiHandle->jobID;

	if ((delib_error_code = tcp_tx_send_rec_data(handle32, DapiHandle->TCPHandle_tx, DapiHandle->jobID, DapiHandle->timeout, DEDITEC_TCP_PACKET_ID_2_CMD_SPECIAL, tx, tx_length, rx, rx_anzahl, rx_cnt)) != DAPI_ERR_NONE)
	{
		if (rx_cnt != NULL)
		{
			*rx_cnt = 0;
		}
	}

	return delib_error_code;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPWriteByte(unsigned long handle32, unsigned long adress, unsigned long value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='B';		// BYTE

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);			
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPWriteWord (unsigned long handle32,unsigned long adress,unsigned long value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='W';		// WORD

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;
	
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 8  ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);			
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPWriteLong (unsigned long handle32,unsigned long adress,unsigned long value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='L';		// Long

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	tx_buffer[tx_cnt++]=(UCHAR) (value >> 24 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 16 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 8  ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);			
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


void DapiTCPWriteLongLong (unsigned long handle32,unsigned long adress, ULONGLONG value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt = 0;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='X';		// LongLong

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	tx_buffer[tx_cnt++]=(UCHAR) (value >> 56 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 48 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 40 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 32 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 24 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 16 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 8  ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);			
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPReadByte (unsigned long handle32, unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='B';		// BYTE

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=1;		// 1 Byte soll empfangen werden

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code == DAPI_ERR_NONE)
	{
		data =((unsigned long) (rx[0]) << 0) & 0x00ff;
		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPReadWord (unsigned long handle32,unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='W';		// WORD

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=2;		// 2 Bytes sollen empfangen werden

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code == DAPI_ERR_NONE)
	{
		data =((unsigned long) (rx[1]) << 8) & 0xff00;
		data|=((unsigned long) (rx[0]) << 0) & 0x00ff;
		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}
	
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPReadLong(unsigned long handle32, unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='L';		// LONG

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=4;		// 4 Bytes sollen empfangen werden

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code == DAPI_ERR_NONE)
	{
		data =((unsigned long) (rx[3]) << 24) & 0xff000000;
		data|=((unsigned long) (rx[2]) << 16) & 0x00ff0000;
		data|=((unsigned long) (rx[1]) << 8 ) & 0x0000ff00;
		data|=((unsigned long) (rx[0]) << 0 ) & 0x000000ff;

		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONGLONG DapiTCPReadLongLong(unsigned long handle32, unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	ULONGLONG data;
	unsigned long tx_cnt = 0;
	unsigned long rx_cnt;
	unsigned long delib_error_code;
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]='X';		// eXtended Mode 32 Adress Bit
	}

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='X';		// LONGLONG

	if(DapiHandle->is_Adr_32_mode == TRUE)
	{
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 24)  & 0xff;
		tx_buffer[tx_cnt++]=(UCHAR) (adress >> 16) & 0xff;
	}
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=8;		// 8 Bytes sollen empfangen werden

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, rx, rx_cnt);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);

	if (delib_error_code == DAPI_ERR_NONE)
	{
		#ifdef __linux__
			data =((ULONGLONG) (rx[7]) << 56) & 0xff00000000000000LL;
			data|=((ULONGLONG) (rx[6]) << 48) & 0x00ff000000000000LL;
			data|=((ULONGLONG) (rx[5]) << 40) & 0x0000ff0000000000LL;
			data|=((ULONGLONG) (rx[4]) << 32) & 0x000000ff00000000LL;
			data|=((ULONGLONG) (rx[3]) << 24) & 0x00000000ff000000LL;
			data|=((ULONGLONG) (rx[2]) << 16) & 0x0000000000ff0000LL;
			data|=((ULONGLONG) (rx[1]) << 8 ) & 0x000000000000ff00LL;
			data|=((ULONGLONG) (rx[0]) << 0 ) & 0x00000000000000ffLL;
		#else
			data =((ULONGLONG) (rx[7]) << 56) & 0xff00000000000000;
			data|=((ULONGLONG) (rx[6]) << 48) & 0x00ff000000000000;
			data|=((ULONGLONG) (rx[5]) << 40) & 0x0000ff0000000000;
			data|=((ULONGLONG) (rx[4]) << 32) & 0x000000ff00000000;
			data|=((ULONGLONG) (rx[3]) << 24) & 0x00000000ff000000;
			data|=((ULONGLONG) (rx[2]) << 16) & 0x0000000000ff0000;
			data|=((ULONGLONG) (rx[1]) << 8 ) & 0x000000000000ff00;
			data|=((ULONGLONG) (rx[0]) << 0 ) & 0x00000000000000ff;
		#endif
		
		return data;
	}
	else 
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPSpecialCmd(unsigned long handle32, unsigned long cmd, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl, unsigned long *rx_cnt)
{
	UCHAR tx_buffer[TX_BUFFER_CP_LENGTH];
	unsigned long tx_cnt=0;
//	unsigned long i;
	unsigned long delib_error_code;
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		DapiSetError(DAPI_ERR_GEN_ILLEGAL_HANDLE, 0, NULL, __FILE__, __LINE__);					
		return 0;
	}


	tx_buffer[tx_cnt++]='Z';		// SpecialCmd
	tx_buffer[tx_cnt++]='C';		// SpecialCmd

	tx_buffer[tx_cnt++]=(UCHAR) (cmd >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (cmd >> 0 ) & 0xff;
	

	if (buffcpy_n(tx_buffer + tx_cnt, sizeof(tx_buffer)-tx_cnt, tx, tx_length) != RETURN_OK)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSpecialCmd");
		sprintf(debug_msg, "Error while buffcpy_n: tx_buffer too small return is %d dec", DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR);
		debug_error(debug_msg);
		// -----

		return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;
	}
	tx_cnt+= tx_length;

	if ((delib_error_code = DapiTCPSendSpecialCommand(handle32, tx_buffer, tx_cnt, rx, rx_anzahl, rx_cnt)) != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}

	return delib_error_code;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPWriteMultipleBytesOLD(ULONG handle32, ULONG adress, ULONG adress_tiefe, ULONG wiederholungen, unsigned char* data, ULONG data_length)
{

	ULONG adr;
	ULONG anz;
	ULONG pos=0;
	ULONGLONG d64;
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}

	debug_set_function_name("DapiTCPWriteMultipleBytesOLD");
	sprintf(debug_msg,"Using OLD Function !!!!");
	debug_print(debug_msg);

	for(anz=0;anz!=wiederholungen;++anz)
	{
		adr=adress;
		
		do 
		{
			if((adr+8) <= (adress+adress_tiefe))
			{
				// 8 Byte

				d64 =  (ULONGLONG) *(data+pos+0);
				d64|= ((ULONGLONG) *(data+pos+1)) << 8;
				d64|= ((ULONGLONG) *(data+pos+2)) << 16;
				d64|= ((ULONGLONG) *(data+pos+3)) << 24;
				d64|= ((ULONGLONG) *(data+pos+4)) << 32;
				d64|= ((ULONGLONG) *(data+pos+5)) << 40;
				d64|= ((ULONGLONG) *(data+pos+6)) << 48;
				d64|= ((ULONGLONG) *(data+pos+7)) << 56;

				DapiWriteLongLongXXX(handle32, adr, d64);
				adr+=8;
				pos+=8;
			}
			else
			{
				DapiWriteByteXXX(handle32, adr, *(data+pos));
				adr++;
				pos++;
			}
		} while(adr < (adress+adress_tiefe));
	}
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPWriteMultipleBytes(ULONG handle32, ULONG address, ULONG address_depth, ULONG increment, unsigned char *data, ULONG data_length)
{
	UCHAR *tx_buffer;
	char debug_msg[200];
	
	unsigned long tx_cnt;
	unsigned long delib_error_code;

	ULONG data_id = 0;
	ULONG i;
	ULONG j;
	ULONG pos;
	
	tx_cnt=0;
	pos=0;
	
	// speicherallokierung fr tx_buffer + check 
	tx_buffer = (UCHAR *) malloc((((address_depth + 3) * increment) + TCP_RO_BIG_TX_HEADER_LENGTH));	// header(8) + wiederholung * (adresstiefe + start_id(1) + data_id(2))
	
	if (tx_buffer == NULL)
	{
		sprintf(debug_msg, "DapiTCPWriteMultipleBytes / malloc() for tx_buffer failed");
		DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}
	
	// tx_buffer ok - daten fllen
	tx_buffer[tx_cnt++]='W';	// COMMAND //Write
	tx_buffer[tx_cnt++]='M';	// WIDTH //Multiple

	// address (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 8 ) & 0xff);	// ADDRESS_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 0 ) & 0xff);	// ADDRESS_BIT_0_7
	
	// address_depth (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 8 ) & 0xff);	// ADDRESS_DEPTH_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 0 ) & 0xff);	// ADDRESS_DEPTH_BIT_0_7
	
	// increment (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 8 ) & 0xff);	// INCREMENT_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 0 ) & 0xff);	// INCREMENT_BIT_0_7
	
	// data start
	for (i=0; i<increment; i++)
	{
		// START_ID_MULTIPLE_BYTE (1byte)
		tx_buffer[tx_cnt++]= DEDITEC_TCP_START_ID_FOR_MULTIPLE_BYTE_DATA;		// START_ID_MULTIPLE_BYTE
		
		// DATA_ID (2byte)
		tx_buffer[tx_cnt++]= (unsigned char) ((i >> 8 ) & 0xff);			// DATA_ID_BIT_8_15
		tx_buffer[tx_cnt++]= (unsigned char) ((i >> 0 ) & 0xff);			// DATA_ID_BIT_0_7
		
		// data
		for (j=0; j<address_depth; j++)
		{
			tx_buffer[tx_cnt++]= (*(data + (pos++)));
		}
	}
	// data end
	
	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, 0, 0);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}

	// speicherplatz von tx_buffer wieder freigeben
	free(tx_buffer);
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPReadMultipleBytes(ULONG handle32, ULONG address, ULONG address_depth, ULONG increment, unsigned char *data, ULONG data_size)
{
	char debug_msg[200];

	UCHAR tx_buffer[2000];
	UCHAR *data_temp;
	unsigned long data_temp_size;
	ULONG data_temp_length;
	unsigned long delib_error_code;
	
	ULONG i;
	ULONG j;
	ULONG pos;
	ULONG tx_cnt;
	
	ULONG m_start_id;
	ULONG m_data_id;
	ULONG m_data_pos;

	DAPI_HANDLE *DapiHandle;
	ULONGLONG handle_ull;

	if(DapiInternHandlePointerGet(handle32, &handle_ull) == RETURN_OK)
	{
		DapiHandle = (DAPI_HANDLE*) DAPI_PTR_SIZE handle_ull;
	}
	else
	{
		return DAPI_ERR_GEN_ILLEGAL_HANDLE;
	}
	
	// check ob der "richtige buffer" gro genug ist	
	if (data_size < (increment * address_depth))
	{
		sprintf(debug_msg, "DapiTCPReadMultipleBytes / Buffer too small - address_depth=%ld * increment=%ld < data_length=%ld", (unsigned long) address_depth, (unsigned long) increment, (unsigned long) data_size);
		DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}	

	// speicherallokierung fr data_temp + check 
	data_temp_size = (((address_depth + 3) * increment) + 1);
	data_temp = (UCHAR *) malloc(data_temp_size);	// +1 weil "sizeof" nicht bei 0 sondern bei 1 anfngt
	
	if (data_temp == NULL)
	{
		sprintf(debug_msg, "DapiTCPReadMultipleBytes / malloc() for temp_buffer failed");
		DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}
	
	data_temp_length = ((address_depth + 3) * increment);//sizeof(data_temp);
	
	// tx_buffer zusammenbauen
	tx_cnt=0;
	pos=0;

	tx_buffer[tx_cnt++]='R';	// COMMAND //Write
	tx_buffer[tx_cnt++]='M';	// WIDTH //Multiple

	// address (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 8 ) & 0xff);	// ADDRESS_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 0 ) & 0xff);	// ADDRESS_BIT_0_7
	
	// address_depth (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 8 ) & 0xff);	// ADDRESS_DEPTH_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 0 ) & 0xff);	// ADDRESS_DEPTH_BIT_0_7
	
	// increment (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 8 ) & 0xff);	// INCREMENT_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 0 ) & 0xff);	// INCREMENT_BIT_0_7
	
	// anfrage an ETH-modul schicken

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommand(handle32, tx_buffer, tx_cnt, data_temp, data_temp_length);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_MULTIPLE, delib_error_code);

	if(delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);					
	}
	
	// data_temp berprfen
	// datenblock = start_id_multiple_byte(1) +  data_id(2) + addr_depth	

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPReadMultipleBytes");

		sprintf(debug_msg,"RECEIVING %lu bytes:", data_temp_length);
		debug_print(debug_msg);

		if (data_temp_length > 0)
		{
			for (i=0; i<data_temp_length; i+=16)
			{
				debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), data_temp, data_temp_size, i, data_temp_length);
				debug_collect(debug_msg);
			}
		}
	#endif
	// -----
	
	for (i=0; i<increment; i++)
	{
		pos = 0;
		m_data_pos = i*(address_depth+3);
		
		// start_id + data_id aus dem buffer filtern
		m_start_id = *(data_temp + m_data_pos + (pos++));
		m_data_id = *(data_temp + m_data_pos + (pos++));
		m_data_id = (m_data_id << 8) | *(data_temp + m_data_pos + (pos++));
		
		// start_id checken
		if (m_start_id != DEDITEC_TCP_START_ID_FOR_MULTIPLE_BYTE_DATA)
		{
			sprintf(debug_msg, "DapiTCPReadMultipleBytes / Wrong START_ID");
			DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

			return 1;			// Error
		}
		
		// data_id checken
		if (m_data_id != i)
		{	
			sprintf(debug_msg, "DapiTCPReadMultipleBytes / Wrong DATA_ID");
			DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

			return 1;			// Error
		}
	}
	
	// daten sind ok - daten in "richtigen buffer" schreiben
	tx_cnt = 0;
	
	for (i=0; i< increment; i++)
	{
		for (j=0; j<address_depth; j++)
		{
			*(data + (tx_cnt++)) = *(data_temp + 3 + j + i*(address_depth+3));
		}
	}
	
	// zum abschluss geschrieben zeichen in "richtigen buffer" mit der errechneten (und bentigten gre vergleichen)
	if (tx_cnt != (increment * address_depth))
	{
		sprintf(debug_msg, "DapiTCPReadMultipleBytes / Buffer copy error");
		DapiSetError(DAPI_ERR_GEN_ERROR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPReadMultipleBytes");

		sprintf(debug_msg,"RETURNING %lu bytes:", data_temp_length);
		debug_print(debug_msg);
		if (tx_cnt > 0)
		{
			for (i=0; i<tx_cnt; i+=16)
			{
				debug_hexdump_line_to_msg(debug_msg, sizeof(debug_msg), data, data_size, i, tx_cnt);
				debug_collect(debug_msg);
			}
		}
	#endif
	// -----
	
	// speicherplatz von data_temp wieder freigeben
	free(data_temp);	
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPReadFifo(ULONG handle32, ULONG type, unsigned char *data, ULONG data_length)
{
	UCHAR tx_buffer[2000];
	ULONG pos;
	ULONG tx_cnt;
	unsigned long rx_data_length;
	unsigned long delib_error_code;

	// tx_buffer zusammenbauen
	tx_cnt=0;
	pos=0;

	tx_buffer[tx_cnt++]='R';	// COMMAND //Write
	tx_buffer[tx_cnt++]='F';	// WIDTH //Multiple

	// type (4byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 24) & 0xff);	// type
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 16) & 0xff);	// type
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 8 ) & 0xff);	// type
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 0 ) & 0xff);	// type
	
	// Anfrage an ETH-modul schicken

	PacketStatisticStart(handle32);
	delib_error_code = DapiTCPSendCommandBig(handle32, tx_buffer, tx_cnt, data, data_length, &rx_data_length);
	PacketStatisticEnd(handle32, PKT_STAT_CMD_TYPE_NORMAL, delib_error_code);
	
	if (delib_error_code != DAPI_ERR_NONE)
	{
		DapiSetError(delib_error_code, 0, NULL, __FILE__, __LINE__);
		rx_data_length=0;
	}
	

	return rx_data_length;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
