//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	deditec_encryption_utils.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2017
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	28.08.2014	Start DEDITEC-History
// 
// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------
//
// dt_encryption_handler
// dt_encrypt_buffer
// dt_decrypt_buffer
//
// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdio.h>
#include <stdlib.h>

#include "../library/embedded/dt_var_defines.h"
#include "../library/delib/delib_error_codes.h"
#include "blowfish.h"

// ------------------------------------
// ------------------------------------
// Defines
// ------------------------------------
// ------------------------------------

#define TYPE_DECRYPT	1
#define TYPE_ENCRYPT	2

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long dt_encryption_handler(unsigned long type, char *key, unsigned long key_length, unsigned char *buffer, unsigned long buffer_length)
{
	BLOWFISH_CTX ctx;
	unsigned long left;
	unsigned long right;
	unsigned long pos;
	unsigned long i;
	
	if (buffer_length % 8 != 0)
	{
//		debug_error("dt_encryption_handler: buffer_length(%ld) % 8 != 0\n", buffer_length);
		return DAPI_ERR_GEN_BUFFER_TOO_SMALL_ERROR;		
	}
	
	Blowfish_Init (&ctx, (unsigned char*) key, key_length);
	
	for (i=0; i<buffer_length; i+=8)
	{		
		pos = i;
		left  = (buffer[pos++] << 24);
		left |= (buffer[pos++] << 16);
		left |= (buffer[pos++] <<  8);
		left |= (buffer[pos++] <<  0);
		
		right  = (buffer[pos++] << 24);
		right |= (buffer[pos++] << 16);
		right |= (buffer[pos++] <<  8);
		right |= (buffer[pos++] <<  0);
		
		switch (type)
		{
			case TYPE_DECRYPT:	Blowfish_Decrypt(&ctx, &left, &right);	break;
			case TYPE_ENCRYPT:	Blowfish_Encrypt(&ctx, &left, &right);	break;
		}

		pos = i;
		buffer[pos++] = (unsigned char)((left >> 24) & 0xff);
		buffer[pos++] = (unsigned char)((left >> 16) & 0xff);
		buffer[pos++] = (unsigned char)((left >>  8) & 0xff);
		buffer[pos++] = (unsigned char)((left >>  0) & 0xff);
		
		buffer[pos++] = (unsigned char)((right >> 24) & 0xff);
		buffer[pos++] = (unsigned char)((right >> 16) & 0xff);
		buffer[pos++] = (unsigned char)((right >>  8) & 0xff);
		buffer[pos++] = (unsigned char)((right >>  0) & 0xff);
	}

	return RETURN_OK;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long dt_decrypt_buffer(char *key, unsigned long key_length, unsigned char *buffer, unsigned long buffer_length)
{
	return dt_encryption_handler(TYPE_DECRYPT, key, key_length, buffer, buffer_length);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long dt_encrypt_buffer(char *key, unsigned long key_length, unsigned char *buffer, unsigned long buffer_length)
{
	return dt_encryption_handler(TYPE_ENCRYPT, key, key_length, buffer, buffer_length);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
