//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	deditec_string_utils.cpp
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2009-2017
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
//
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History
// CC	23.07.2013	Added toupperchar_n
// 
// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------
//
//	String					Buffer
// -------------------------------------------
// strcat_n					buffcat_n
// strcpy_n					buffcpy_n
// strcmp_n					buffcmp_n
// 
// cuttext_n

// make_line_clean
// strcpy_n***************
// strcat_max
// unescape_str
// cutontext*****************
// cuttext*****************


// strcmp_n********************
// strcasecmp_n
// debug_hexdump_buffer_to_string
// str_replace_char_n
// toupperstr_n
// toupperchar_n

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdio.h>
#include <stdlib.h>

#include "deditec_convert_functions.h"
#include "../embedded/dt_var_defines.h"

// ------------------------------------
// ------------------------------------
// Defines
// ------------------------------------
// ------------------------------------

#define ERROR 100
#define LINE_LENGTH 1000

#define DEST_BUFF_WR(x, y) if(x < dest_buffer_size) {dest_buffer[x] = source_buffer[y];} else {ret = 1;};

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long strlen_n(char *buffer, unsigned long buffer_size)
{
	unsigned long pos = 0;
	bool found = FALSE;
	
	do
	{
		if (*(buffer + pos) == '\0')
		{
			found = TRUE;
		}
		else
		{
			pos++;
		}
	
	} while((pos<buffer_size) && (found!=TRUE));
	
	return pos;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long strcat_n(	char *dest_buffer,   unsigned long dest_buffer_size, 
						char *source_buffer, unsigned long source_buffer_size)
{
	unsigned long ret = 0;
	unsigned long dest_pos = 0;
	unsigned long source_pos = 0;

	while(dest_buffer[dest_pos] != 0)
	{
		dest_pos++;
	}

	while(source_buffer[source_pos] != 0)
	{
		DEST_BUFF_WR(dest_pos, source_pos);
		dest_pos++;
		source_pos++;
	}

	if(dest_pos < dest_buffer_size)
	{
		dest_buffer[dest_pos] = '\0';
	}
	else
	{
		ret = 1;
	}

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long strcpy_n(	char *dest_buffer,   unsigned long dest_buffer_size, 
						char *source_buffer, unsigned long source_buffer_size)
{
	unsigned long ret = 0;
	unsigned long pos = 0;

	unsigned long max;
	char c;

	if(dest_buffer_size > 0)
	{
		dest_buffer[0] = 0;
	}

	max = dest_buffer_size;
	if (max > source_buffer_size) max = source_buffer_size;

	if(max == 0)
	{
		ret = 1;		//Fehler
	}

	pos=0;

	do
	{
		if(pos < max)
		{
			c = source_buffer[pos];
			dest_buffer[pos] = c;
		}
		else
		{
			c = 0;
			ret = 1;
		}

		pos++;

	} while((c != 0) && (pos < max));

	if(pos == dest_buffer_size)
	{
		--pos;
		ret = 1;
	}

	dest_buffer[pos] = '\0';		// Terminate

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long cuttext_n(char *right_string, 	unsigned long right_string_size,
					    char *left_string,  	unsigned long left_string_size,
					    char *pattern,		  	unsigned long pattern_length)
{
	unsigned long pos_right_string;
	unsigned long pos_left_string;
	unsigned long pos_pattern;
	unsigned long found;

	pos_right_string = 0;
	pos_left_string = 0;
	pos_pattern = 0;
	found = FALSE;

	while ((pos_right_string < right_string_size) && (pos_right_string < strlen_n(right_string, -1)))
	{
		if ( *(right_string + pos_right_string) == *(pattern + pos_pattern) )
		{
			pos_right_string++;
			pos_pattern++;

			if (pos_pattern == pattern_length)
			{
				found = TRUE;

				if (left_string_size - 1 < pos_right_string - pos_pattern)
				{
					return ERROR;
				}
				strcpy_n(left_string, left_string_size, right_string, pos_right_string - pos_pattern);
				strcpy_n(right_string, right_string_size, &right_string[pos_right_string], strlen_n(right_string, -1) - pos_right_string );

				return found;
			}
		}
		else
		{
			pos_pattern = 0;

			if ( *(right_string + pos_right_string) == *(pattern + pos_pattern) )
			{
			pos_pattern++;
			}
			pos_right_string++;
		}
	}
	
	return found;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long make_line_clean(char *string, unsigned long string_size)
{
	unsigned long i;
	unsigned long i_destination;
	unsigned long ret;

	ret = 0;

	i = 0;
	while (string[i] != 0 && i <= string_size - 1)
	{
		i++;
	}
	
	i--;
	
	while (string[i] == 10  || string[i] == 13  || string[i] == ' ')
	{
		string[i] = 0;
		i--;
		ret = 1;
	}

   // removing beginnig spaces
	i = 0;
	while(string[i] == ' ')
	{
		i++;
	}
	
	i_destination = 0;
	
	if (i > 0) 
	{
		while (i <= string_size - 1)
		{
			string[i_destination] = string[i];
			i++;
			i_destination++;
			ret = 1;
		}
	}
	
	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long strcat_max (char *dest_buffer,   unsigned long dest_buffer_size,
                          char *source_buffer, unsigned long amount_of_chars)
{
	unsigned long ret;
	unsigned long pos_dest_buffer;

	ret = 0;
	pos_dest_buffer = 0;

	while(*(dest_buffer + pos_dest_buffer) != 0 && pos_dest_buffer < dest_buffer_size - 1)
	{
		pos_dest_buffer++;
	}

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned char unescape_str(unsigned char *escape,	unsigned long escape_length,
						   unsigned char *unescape, unsigned long unescape_length)
{
	unsigned long pos_escape;
	unsigned long pos_unescape;
	unsigned char asc_p1;
	unsigned char asc_p2;

	pos_unescape = 0;
	*(unescape + pos_unescape) = 0;


	for (pos_escape = 0; ((*(escape + pos_escape) != 0) && (pos_escape < escape_length)); pos_escape++)
	{
		if (*(escape + pos_escape) == '%')
		{
			if (pos_escape + 2 >= escape_length)
			{
				*(unescape + pos_unescape) = 0;
				
				return FALSE;
			}
	
			++pos_escape;

			asc_p1 = *(escape + pos_escape++);
			asc_p2 = *(escape + pos_escape);


			if (!((asc_p1 >= '0' && asc_p1 <= '9') ||
			(asc_p1 >= 'A' && asc_p1 <= 'F') ||
			(asc_p1 >= 'a' && asc_p1 <= 'f')) )
			{
				*(unescape + pos_unescape) = 0;
				return FALSE;
			}

			if (!((asc_p2 >= '0' && asc_p2 <= '9') ||
			(asc_p2 >= 'A' && asc_p2 <= 'F') ||
			(asc_p2 >= 'a' && asc_p2 <= 'f')) )
			{
				*(unescape + pos_unescape) = 0;
				return FALSE;
			}

			*(unescape + pos_unescape)  = (unsigned char) ((asc2hex(asc_p1) << 4) | asc2hex(asc_p2));

			pos_unescape++;
		}
		else if (*(escape + pos_escape) == '+')
		{
			*(unescape + pos_unescape) = ' ';
			pos_unescape++;
		}
		else
		{
			*(unescape + pos_unescape) = *(escape + pos_escape);
			pos_unescape++;
		}
	}

	*(unescape + pos_unescape) = 0;
	return TRUE;
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long strcmp_n (char *needle, unsigned long needle_length, char *heystack, unsigned long heystack_length)
{
	unsigned long i;
	unsigned long min;
	
	min = needle_length < heystack_length ? needle_length : heystack_length;

	for (i = 0; i < min; i++)
	{
		if (needle[i] != heystack[i])
		{
			return 1;
		}
		if (needle[i] == 0)
		{
			return 0;
		}
	}
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long strcasecmp_n (char *needle, unsigned long needle_length, char *heystack, unsigned long heystack_length)
{
	unsigned long i;
	unsigned long min;
	char nc, hc;

	min = needle_length < heystack_length ? needle_length : heystack_length;

	for (i = 0; i < min; i++)
	{
		nc = needle[i];
		hc = heystack[i];
		nc = ((nc >= 'A') && (nc <= 'Z')) ? (nc - 'A' + 'a') : nc;
		hc = ((hc >= 'A') && (hc <= 'Z')) ? (hc - 'A' + 'a') : hc;
		if (nc != hc)
		{
			return 1;
		}
	}
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long debug_hexdump_buffer_to_string(unsigned char *src_buffer,  unsigned long src_buffer_len, 
													  char *dest_buffer, unsigned long dest_buffer_len)
{
	unsigned long i;
	char temp_msg[20];	
	char msg_is_capped;

	// check destination
	if (dest_buffer_len == 0)
	{
		return 1;
	}
	
	// dest_buffer leeren + cap_flag init
	*dest_buffer = 0;
	msg_is_capped = 0;
	
	// check source
	if (src_buffer_len == 0)
	{
		strcpy_n(temp_msg, sizeof(temp_msg), "NO DATA", sizeof("NO DATA"));
		if((strlen_n(dest_buffer, dest_buffer_len) + strlen_n(temp_msg, sizeof(temp_msg))) < dest_buffer_len)			// Passt in den Dest-Buffer (incl. 0-Terminierung)
		{
			strcat_n(dest_buffer, dest_buffer_len, temp_msg, strlen_n(temp_msg, sizeof(temp_msg)));
		}
	}
	else
	{
		// Schleife
		i=0;
		do
		{
			if((strlen_n(dest_buffer, sizeof(dest_buffer)) + 3 + 5) < dest_buffer_len)			// Passt in den Dest-Buffer (incl. 0-Terminierung)
			{
				sprintf(temp_msg, "%02x ", *(src_buffer+i)),
				strcat_n(dest_buffer, dest_buffer_len, temp_msg, strlen_n(temp_msg, sizeof(temp_msg)));
			}
			else
			{
				strcpy_n(temp_msg, sizeof(temp_msg), "CAP", sizeof("CAP"));
				strcat_n(dest_buffer, dest_buffer_len, temp_msg, strlen_n(temp_msg, sizeof(temp_msg)));
				msg_is_capped = 1;
			}
			
			++i;
		} while ( (i<src_buffer_len) && (msg_is_capped==0) );
	}

	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long str_replace_char_n(	char *src_buffer, 	unsigned long src_buffer_length, 
									char old_char,	char new_char)
{
	unsigned long pos;
	
	for (pos=0; pos!=src_buffer_length; pos++)
	{
		/// TODO warning: comparison between pointer and integer
		if (*(src_buffer + pos) == old_char)
		{
			/// TODO  warning: assignment makes integer from pointer without a cast
			*(src_buffer + pos) = new_char;
		}
	}
	
	return RETURN_OK;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long toupperstr_n(char *src_buffer, unsigned long src_buffer_length)
{
	unsigned long pos;
	
	for (pos=0; pos!=src_buffer_length; pos++)
	{
		if ((src_buffer[pos] >= 'a') && src_buffer[pos] <= 'z')
		{
			src_buffer[pos] -= 32;
		}
	}
	
	return RETURN_OK;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

char toupperchar_n(char character)
{
	if( (character >= 'a') && (character <= 'z') ) 
	{
		return (character-32);
	}
	else
	{
		return character;
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
